"""Data loader for beamline 5-2 at SSRL."""

import os
import re
import datetime
import numpy as np
import numpy.typing as npt
import xarray as xr
import h5netcdf
import pandas as pd


import erlab.io.utilities
from erlab.io.dataloader import LoaderBase


class SSRL52Loader(LoaderBase):
    name: str = "ssrl"
    aliases: list[str] = ["ssrl52", "bl5-2"]

    name_map: dict[str, str] = {
        "eV": "Kinetic Energy",
        "alpha": "ThetaX",
        "beta": ["ThetaY", "YDeflection", "DeflectionY"],
        "delta": ["A", "a"],  # azi
        "chi": ["T", "t"],  # polar
        "xi": ["F", "f"],  # tilt
        "x": "X",
        "y": "Y",
        "z": "Z",
        "hv": ["BL_energy", "BL_photon_energy"],
        "temp_sample": ["TB", "sample_stage_temperature"],
        "sample_workfunction": "WorkFunction",
    }
    coordinate_attrs: tuple[str, ...] = (
        "beta",
        "delta",
        "chi",
        "xi",
        "hv",
        "x",
        "y",
        "z",
    )
    additional_attrs: dict[str, str | int | float] = {
        "configuration": 3,
        "sample_workfunction": 4.5,
    }

    always_single: bool = True
    skip_validate: bool = True

    def load_single(self, file_path: str | os.PathLike) -> xr.DataArray:
        with h5netcdf.File(file_path, mode="r", phony_dims="sort") as ncf:
            attrs = dict(ncf.attrs)
            compat_mode = "data" in ncf.groups  # Compatibility with older data
            for k, v in ncf.groups.items():
                ds = xr.open_dataset(xr.backends.H5NetCDFStore(v, autoclose=True))

                if k.casefold() == "Beamline".casefold():
                    attrs[k] = ds.attrs
                    hv = ds.attrs.get("energy", None)
                    hv = ds.attrs.get("photon_energy", hv)
                    if hv is not None:
                        attrs["hv"] = hv

                    attrs["polarization"] = ds.attrs.get("polarization")

                else:
                    # Merge group attributes
                    attrs = attrs | ds.attrs

                if k.casefold() == "Data".casefold():
                    if compat_mode:
                        if "exposure" in ds.variables:
                            ds = ds.rename_vars(counts="spectrum", exposure="time")
                        else:
                            ds = ds.rename_vars(counts="spectrum")
                    elif "Time" in ds.variables:
                        ds = ds.rename_vars(Count="spectrum", Time="time")
                    else:
                        ds = ds.rename_vars(Count="spectrum")

                    # List of dicts containing scale and label info for each axis
                    axes: list[dict[str, float | int | str]] = [
                        dict(v.groups[g].attrs) for g in v.groups
                    ]

                    for i, ax in enumerate(axes):
                        # Unify case for compatibility with old data
                        axes[i] = {name.lower(): val for name, val in ax.items()}

                    # Apply dim labels
                    data = ds.rename_dims(
                        {f"phony_dim_{i}": ax["label"] for i, ax in enumerate(axes)}
                    ).load()

                    # Apply coordinates
                    for i, ax in enumerate(axes):
                        if compat_mode:
                            cnt = v.dimensions[f"phony_dim_{i}"].size
                        else:
                            cnt = ax["count"]
                        mn, mx = (
                            ax["offset"],
                            ax["offset"] + (cnt - 1) * ax["delta"],
                        )
                        data = data.assign_coords(
                            {ax["label"]: np.linspace(mn, mx, cnt)}
                        )

        if "time" in data.variables:
            # Normalize by dwell time
            data = data["spectrum"] / data["time"]
        else:
            data = data["spectrum"]

        data = data.assign_attrs(attrs)
        return self.process_keys(data)

    def identify(
        self,
        num: int,
        data_dir: str | os.PathLike,
        zap: bool = False,
    ) -> tuple[list[str], dict[str, npt.NDArray[np.float64]]]:
        if zap:
            target_files = erlab.io.utilities.get_files(
                data_dir, extensions=(".h5",), contains="zap"
            )
        else:
            target_files = erlab.io.utilities.get_files(
                data_dir, extensions=(".h5",), notcontains="zap"
            )

        for file in target_files:
            match = re.match(r"(.*?)_" + str(num).zfill(4) + r".h5", file)
            if match is not None:
                return [file], {}

        raise FileNotFoundError(f"No files found for scan {num} in {data_dir}")

    # def post_process(
    #     self, data: xr.DataArray | xr.Dataset
    # ) -> xr.DataArray | xr.Dataset:
    #     data = super().post_process(data)

    #     if "eV" in data.coords:
    #         data = data.assign_coords(eV=-data.eV.values)

    #     return data

    def load_zap(self, identifier, data_dir):
        return self.load(identifier, data_dir, zap=True)

    def generate_summary(
        self, data_dir: str | os.PathLike, exclude_zap: bool = False
    ) -> pd.DataFrame:
        files: dict[str, str] = {}

        if exclude_zap:
            target_files = erlab.io.utilities.get_files(
                data_dir, extensions=(".h5",), notcontains="zap"
            )
        else:
            target_files = erlab.io.utilities.get_files(data_dir, extensions=(".h5",))

        for pth in target_files:
            base_name = os.path.splitext(os.path.basename(pth))[0]
            files[base_name] = pth

        summary_attrs: dict[str, str] = {
            "Type": "Description",
            "Lens Mode": "LensModeName",
            "Region": "RegionName",
            "T(K)": "temp_sample",
            "Pass E": "PassEnergy",
            "Polarization": "polarization",
            "hv": "hv",
            # "Entrance Slit": "Entrance Slit",
            # "Exit Slit": "Exit Slit",
            "x": "x",
            "y": "y",
            "z": "z",
            "polar": "chi",
            "tilt": "xi",
            "azi": "delta",
            "DA": "beta",
        }

        cols = ["Time", "File Name", *summary_attrs.keys()]

        data_info = []

        for name, path in files.items():
            data = self.load(path)

            data_info.append(
                [datetime.datetime.fromtimestamp(data.attrs["CreationTimeStamp"]), name]
            )

            for k, v in summary_attrs.items():
                try:
                    val = data.attrs[v]
                except KeyError:
                    try:
                        val = data.coords[v].values
                        if val.size == 1:
                            val = val.item()
                    except KeyError:
                        val = ""

                if k == "Pass E":
                    val = round(val)

                elif k == "Polarization":
                    if np.iterable(val):
                        val = np.round(np.asarray(val), 3).astype(float)
                    else:
                        val = [float(np.round(val, 3))]
                    val = [
                        {0.0: "LH", 0.5: "LV", 0.25: "RC", -0.25: "LC"}.get(v, v)
                        for v in val
                    ]

                    if len(val) == 1:
                        val = val[0]

                data_info[-1].append(val)

            del data

        return (
            pd.DataFrame(data_info, columns=cols)
            .sort_values("Time")
            .set_index("File Name")
        )



"""Models for fitting data."""

__all__ = [
    "ExtendedAffineBroadenedFD",
    "StepEdgeModel",
    "PolynomialModel",
    "MultiPeakModel",
    "FermiEdge2dModel",
]


import lmfit
import numba
import numpy as np
import numpy.typing as npt
import scipy.ndimage
import xarray as xr
from arpes.fits import XModelMixin

from erlab.analysis.fit.functions import (
    fermi_dirac_linbkg_broad,
    step_linbkg_broad,
    PolyFunc,
    MultiPeakFunction,
    FermiEdge2dFunc,
)


@numba.njit("f8[:,:](f8[:], i8)", cache=True)
def _coeff_mat(x, deg):
    mat_ = np.zeros(shape=(x.shape[0], deg + 1), dtype=np.float64)
    const = np.ones_like(x)
    mat_[:, 0] = const
    mat_[:, 1] = x
    if deg > 1:
        for n in range(2, deg + 1):
            mat_[:, n] = x**n
    return mat_


@numba.njit("f8[:](f8[:,:], f8[:])", cache=True)
def _fit_x(a, b):
    # linalg solves ax = b
    det_ = np.linalg.lstsq(a, b)[0]
    return det_


@numba.njit("f8[:](f8[:], f8[:], i8)", cache=True)
def fit_poly_jit(x: npt.NDArray[np.float64], y: npt.NDArray[np.float64], deg: np.int64):
    """`numba`-accelerated polynomial fitting.

    Parameters
    ----------
    x
        The x-coordinates of the data points.
    y
        The y-coordinates of the data points.
    deg
        The degree of the polynomial to fit.

    Returns
    -------
    p : ndarray
        1D array of polynomial coefficients (including coefficients equal to zero) from
        the constant term to the highest degree.

    """
    a = _coeff_mat(x, deg)
    p = _fit_x(a, y)
    return p


def fit_edges_linear(x, data, len_fit) -> tuple[float, float, float, float]:
    """Fit :math:`y = mx + n` to each edge of the data.

    Parameters
    ----------
    x
        The x-coordinates of the data points.
    data
        The y-coordinates of the data points.
    len_fit
        The number of data points to use for fitting the edges.

    Returns
    -------
    n0, m0
        The coefficients of the linear fit for the left edge.
    n1, m1
        The coefficients of the linear fit for the right edge.
    """
    n0, m0 = fit_poly_jit(
        np.array(x[:len_fit], dtype=np.float64),
        np.array(data[:len_fit], dtype=np.float64),
        deg=1,
    )
    n1, m1 = fit_poly_jit(
        np.array(x[-len_fit:], dtype=np.float64),
        np.array(data[-len_fit:], dtype=np.float64),
        deg=1,
    )
    return n0, m0, n1, m1


class ExtendedAffineBroadenedFD(XModelMixin):
    """
    Fermi-dirac function with linear background above and below the fermi level,
    convolved with a gaussian kernel.
    """

    guess_dataarray = True

    @staticmethod
    def LinearBroadFermiDirac(
        x,
        center=0,
        temp=30,
        resolution=0.02,
        back0=1,
        back1=0,
        dos0=1,
        dos1=0,
    ):
        return fermi_dirac_linbkg_broad(
            x, center, temp, resolution, back0, back1, dos0, dos1
        )

    def __init__(
        self, independent_vars=("x",), prefix="", missing="raise", name=None, **kwargs
    ):
        """Defer to lmfit for initialization."""
        kwargs.update(
            {"prefix": prefix, "missing": missing, "independent_vars": independent_vars}
        )
        super().__init__(self.LinearBroadFermiDirac, **kwargs)
        self.set_param_hint("temp", min=0.0)
        self.set_param_hint("resolution", min=0.0)

    def guess(self, data, x, **kwargs):
        """Make some heuristic guesses."""
        pars = self.make_params()

        len_fit = max(round(len(x) * 0.05), 10)
        dos0, dos1, back0, back1 = fit_edges_linear(x, data, len_fit)
        efermi = x[
            np.argmin(np.gradient(scipy.ndimage.gaussian_filter1d(data, 0.2 * len(x))))
        ]

        temp = 30
        if isinstance(data, xr.DataArray):
            try:
                temp = data.attrs["temp_sample"]
            except KeyError:
                pass

        pars[f"{self.prefix}center"].set(value=efermi, min=x.min(), max=x.max())
        pars[f"{self.prefix}back0"].set(value=back0)
        pars[f"{self.prefix}back1"].set(value=back1)
        pars[f"{self.prefix}dos0"].set(value=dos0)
        pars[f"{self.prefix}dos1"].set(value=dos1)
        pars[f"{self.prefix}temp"].set(value=temp)
        pars[f"{self.prefix}resolution"].set(value=0.02)

        return lmfit.models.update_param_vals(pars, self.prefix, **kwargs)

    __init__.doc = lmfit.models.COMMON_INIT_DOC
    guess.__doc__ = lmfit.models.COMMON_GUESS_DOC


class StepEdgeModel(XModelMixin):
    guess_dataarray = True

    def __init__(
        self, independent_vars=("x",), prefix="", missing="raise", name=None, **kwargs
    ):
        """Defer to lmfit for initialization."""
        kwargs.update(
            {"prefix": prefix, "missing": missing, "independent_vars": independent_vars}
        )
        super().__init__(step_linbkg_broad, **kwargs)
        self.set_param_hint("sigma", min=0.0)

    def guess(self, data, x, **kwargs):
        """Make some heuristic guesses."""
        pars = self.make_params()

        len_fit = max(round(len(x) * 0.05), 10)
        dos0, dos1, back0, back1 = fit_edges_linear(x, data, len_fit)
        efermi = x[
            np.argmin(np.gradient(scipy.ndimage.gaussian_filter1d(data, 0.2 * len(x))))
        ]

        pars[f"{self.prefix}center"].set(value=efermi)
        pars[f"{self.prefix}back0"].set(value=back0)
        pars[f"{self.prefix}back1"].set(value=back1)
        pars[f"{self.prefix}dos0"].set(value=dos0)
        pars[f"{self.prefix}dos1"].set(value=dos1)
        pars[f"{self.prefix}sigma"].set(value=0.02)

        return lmfit.models.update_param_vals(pars, self.prefix, **kwargs)

    __init__.doc = lmfit.models.COMMON_INIT_DOC
    guess.__doc__ = lmfit.models.COMMON_GUESS_DOC


class PolynomialModel(XModelMixin):
    def __init__(self, degree=9, **kwargs):
        kwargs.setdefault("name", f"Poly{degree}")
        super().__init__(PolyFunc(degree), **kwargs)

    def guess(self, data, x=None, **kwargs):
        pars = self.make_params()
        if x is None:
            pars["c0"].set(value=data.mean())
            for i in range(1, self.func.degree + 1):
                pars[f"{self.prefix}c{i}"].set(value=0.0)
        else:
            out = np.polyfit(x, data, self.func.degree)
            for i, coef in enumerate(out[::-1]):
                pars[f"{self.prefix}c{i}"].set(value=coef)
        return lmfit.models.update_param_vals(pars, self.prefix, **kwargs)

    __init__.doc = lmfit.models.COMMON_INIT_DOC
    guess.__doc__ = lmfit.models.COMMON_GUESS_DOC


class MultiPeakModel(XModelMixin):
    """Model for fitting multiple Gaussian or Lorentzian peaks.

    Most input parameters are passed to the
    `erlab.analysis.fit.functions.MultiPeakFunction` constructor.
    """

    def __init__(
        self,
        npeaks: int = 1,
        peak_shapes: list[str] | str | None = None,
        fd: bool = True,
        convolve: bool = True,
        independent_vars=None,
        param_names=None,
        nan_policy="raise",
        prefix="",
        name=None,
        **kws,
    ):
        super().__init__(
            MultiPeakFunction(
                npeaks, peak_shapes=peak_shapes, fd=fd, convolve=convolve
            ),
            independent_vars,
            param_names,
            nan_policy,
            prefix,
            name,
            **kws,
        )

        if self.func.convolve:
            self.set_param_hint("resolution", min=0.0)

    def guess(self, data, x=None, **kwargs):
        pars = self.make_params()
        # !TODO: better guesses
        if self.func.fd:
            pars[f"{self.prefix}offset"].set(value=data[x >= 0].mean())

        poly1 = PolynomialModel(1).guess(data, x)
        pars[f"{self.prefix}lin_bkg"].set(poly1["c1"].value)
        pars[f"{self.prefix}const_bkg"].set(poly1["c0"].value)

        # for i, func in enumerate(self.func.peak_funcs):
        # self.func.peak_argnames

        xrange = x.max() - x.min()

        for i in range(self.func.npeaks):  # Number of peaks
            pars[f"{self.prefix}p{i}_center"].set(value=0.0)
            pars[f"{self.prefix}p{i}_height"].set(value=data.mean())
            pars[f"{self.prefix}p{i}_width"].set(value=0.1 * xrange)

        return lmfit.models.update_param_vals(pars, self.prefix, **kwargs)

    guess.__doc__ = lmfit.models.COMMON_GUESS_DOC


class FermiEdge2dModel(XModelMixin):
    r"""A 2D model for a polynomial Fermi edge with a linear density of states.

    The model function can be written as

    .. math::

        I = \left\{(a\omega + b)\left[1 + \exp\left(\frac{\omega - \sum_{i = 0}^{n} c_i
        \alpha^i}{k_B T}\right)\right]^{-1} + c\right\}\otimes g(\sigma)

    for a :math:`n` th degree polynomial edge with coefficients :math:`c_i` with a
    linear density of states described by :math:`a\omega+b` with a constant background
    :math:`c` convolved with a gaussian, where :math:`\omega` is the binding energy and
    :math:`\alpha` is the detector angle.

    """

    n_dims = 2
    dimension_order = ["eV", "alpha"]
    guess_dataarray = True
    fit_flat = True

    def __init__(
        self,
        degree=2,
        independent_vars=("eV", "alpha"),
        prefix="",
        nan_policy="raise",
        **kwargs,
    ):
        kwargs.update(
            {
                "prefix": prefix,
                "nan_policy": nan_policy,
                "independent_vars": independent_vars,
            }
        )
        super().__init__(FermiEdge2dFunc(degree), **kwargs)
        self.name = f"FermiEdge2dModel (deg {degree})"

    def guess(self, data, eV, alpha, negative=False, **kwargs):
        pars = self.make_params()
        for i in range(self.func.poly.degree + 1):
            pars[f"{self.prefix}c{i}"].set(value=0.0)

        avg_edc = data.mean("alpha").values
        len_fit = max(round(len(eV) * 0.05), 10)
        dos0, dos1 = fit_poly_jit(
            np.asarray(eV[:len_fit], dtype=np.float64),
            np.asarray(avg_edc[:len_fit], dtype=np.float64),
            deg=1,
        )
        pars[f"{self.prefix}const_bkg"].set(value=dos0)
        pars[f"{self.prefix}lin_bkg"].set(value=dos1)
        pars[f"{self.prefix}temp"].set(value=data.attrs["temp_sample"])

        return lmfit.models.update_param_vals(pars, self.prefix, **kwargs)

    def guess_fit(self, *args, **kwargs):
        return super().guess_fit(*args, **kwargs)

    __init__.__doc__ = lmfit.models.COMMON_INIT_DOC.replace("['x']", "['eV', 'alpha']")
    guess.__doc__ = lmfit.models.COMMON_GUESS_DOC



"""Extensions to mplot3d."""

import matplotlib.collections
import matplotlib.patches
import mpl_toolkits.mplot3d
import mpl_toolkits.mplot3d.art3d
import mpl_toolkits.mplot3d.proj3d
import numpy as np


class FancyArrow3D(matplotlib.patches.FancyArrow):
    def __init__(self, x, y, z, dx, dy, dz, **kwargs):
        super().__init__(0, 0, 0, 0, **kwargs)
        posA, posB = (x, y, z), (x + dx, y + dy, z + dz)
        self._verts3d = tuple((posA[i], posB[i]) for i in range(3))

    def do_3d_projection(self, renderer=None):
        xs3d, ys3d, zs3d = self._verts3d
        xs, ys, zs = mpl_toolkits.mplot3d.proj3d.proj_transform(
            xs3d, ys3d, zs3d, self.axes.M
        )
        self.set_data(x=xs[0], y=ys[0], dx=xs[1] - xs[0], dy=ys[1] - ys[0])
        return np.min(zs)


class FancyArrowPatch3D(matplotlib.patches.FancyArrowPatch):
    def __init__(self, posA, posB, *args, **kwargs):
        super().__init__((0, 0), (0, 0), *args, **kwargs)

        self._verts3d = tuple((posA[i], posB[i]) for i in range(3))

    def do_3d_projection(self, renderer=None):
        xs3d, ys3d, zs3d = self._verts3d
        xs, ys, zs = mpl_toolkits.mplot3d.proj3d.proj_transform(
            xs3d, ys3d, zs3d, self.axes.M
        )
        self.set_positions((xs[0], ys[0]), (xs[1], ys[1]))
        return np.min(zs)


# arbitrary patch placing
# improved from https://stackoverflow.com/q/18228966


def _transform_zdir(zdir):
    zdir = mpl_toolkits.mplot3d.art3d.get_dir_vector(zdir)
    zn = zdir / np.linalg.norm(zdir)

    cos_angle = zn[2]
    sin_angle = np.linalg.norm(zn[:2])
    if sin_angle == 0:
        return np.sign(cos_angle) * np.eye(3)

    d = np.array((zn[1], -zn[0], 0))
    d /= sin_angle
    ddt = np.outer(d, d)
    skew = np.array([[0, 0, -d[1]], [0, 0, d[0]], [d[1], -d[0], 0]], dtype=np.float64)
    return ddt + cos_angle * (np.eye(3) - ddt) + sin_angle * skew


def set_3d_properties(self, verts, zs=0, zdir="z"):
    zs = np.broadcast_to(zs, len(verts))
    self._segment3d = np.asarray(
        [
            (*np.dot(_transform_zdir(zdir), (x, y, 0)), 0, 0, z)
            for ((x, y), z) in zip(verts, zs)
        ]
    )


def pathpatch_translate(pathpatch, delta):
    pathpatch._segment3d += np.asarray(delta)


def to_3d(pathpatch, z=0.0, zdir="z", delta=(0, 0, 0)):
    if not hasattr(pathpatch.axes, "get_zlim"):
        raise ValueError("Axes projection must be 3D")
    mpl_toolkits.mplot3d.art3d.pathpatch_2d_to_3d(pathpatch, z=z, zdir=zdir)
    pathpatch.translate(delta)
    return pathpatch


mpl_toolkits.mplot3d.art3d.Patch3D.set_3d_properties = set_3d_properties
mpl_toolkits.mplot3d.art3d.Patch3D.translate = pathpatch_translate
# matplotlib.patches.Patch.to_3d = to_3d



"""Functions related to masking.

Polygon masking is adapted from the `CGAL
<https://doc.cgal.org/5.3.2/Polygon/index.html>`_ library. More information on
point-in-polygon strategies can be found in Ref. :cite:p:`Schirra2008`.

.. currentmodule:: erlab.analysis.mask

Modules
=======

.. autosummary::
   :toctree:

   polygon

"""

__all__ = [
    "mask_with_polygon",
    "polygon_mask",
    "polygon_mask_points",
    "mask_with_hex_bz",
    "hex_bz_mask_points",
]

import numpy as np
import numba
import xarray as xr

from erlab.analysis.mask import polygon


def mask_with_polygon(arr, vertices, dims=("kx", "ky"), invert=False):
    mask = xr.DataArray(
        polygon_mask(
            vertices.astype(np.float64), *(arr[d].values for d in dims), invert=invert
        ),
        dims=dims,
        coords={d: arr.coords[d] for d in dims},
    )
    return arr.where(mask == 1)


@numba.njit(parallel=True, cache=True)
def polygon_mask(vertices, x, y, invert=False):
    shape = (len(x), len(y))
    mask = np.empty(shape, dtype=np.int16)
    vertices = np.concatenate((vertices, vertices[0].reshape(1, -1)))

    for i in numba.prange(shape[0]):
        for j in range(shape[1]):
            mask[i, j] = polygon.bounded_side_bool(vertices, (x[i], y[j]))
    if invert:
        return ~mask
    else:
        return mask


@numba.njit(parallel=True, cache=True)
def polygon_mask_points(vertices, x, y, invert=False):
    if len(x) != len(y):
        raise ValueError
    mask = np.empty(len(x), dtype=np.int16)
    vertices = np.concatenate((vertices, vertices[0].reshape(1, -1)))
    for i in numba.prange(len(x)):
        mask[i] = polygon.bounded_side_bool(vertices, (x[i], y[i]))
    if invert:
        return ~mask
    else:
        return mask


def mask_with_hex_bz(kxymap: xr.DataArray, a=3.54, rotate=0, invert=False):
    """Returns map masked with a hexagonal BZ."""

    if "kx" in kxymap.dims or "qx" in kxymap.dims:
        dims = ("kx", "ky")

    d = 2 * np.pi / (a * 3)
    ang = rotate + np.array([0, 60, 120, 180, 240, 300])
    vertices = np.array(
        [[2 * d * np.cos(t), 2 * d * np.sin(t)] for t in np.deg2rad(ang)]
    )
    return mask_with_polygon(kxymap, vertices, dims, invert=invert)


def hex_bz_mask_points(
    x, y, a=3.54, rotate=0, offset=None, reciprocal=False, invert=False
):
    """Returns a mask for given points."""
    if offset is None:
        offset = (0, 0)
    if reciprocal:
        d = 2 * np.pi / (a * 3)
    else:
        d = a
    ang = rotate + np.array([0, 60, 120, 180, 240, 300])
    vertices = np.array(
        [
            [2 * d * np.cos(t) + offset[0], 2 * d * np.sin(t) + offset[1]]
            for t in np.deg2rad(ang)
        ]
    )
    return polygon_mask_points(vertices, x, y, invert)



"""
This module provides a fast, parallelized version of nanmean that supports multiple axes
based on numba. Enables efficient real-time multidimensional binning.

"""

__all__ = ["fast_nanmean"]

import numpy as np
import numpy.typing as npt
from collections.abc import Iterable
import numba
import numba.typed
import numba.core.registry
import numbagg


@numba.njit(cache=True)
def _nanmean_all(a: npt.NDArray[np.float32 | np.float64]) -> np.float64:
    return np.nanmean(a)


@numba.njit(cache=True, parallel=True)
def _nanmean_2_0(a: npt.NDArray[np.float32 | np.float64]) -> npt.NDArray[np.float64]:
    n = a.shape[1]
    output = np.empty(n, dtype=np.float64)
    for i in numba.prange(n):
        output[i] = np.nanmean(a[:, i])
    return output


@numba.njit(cache=True, parallel=True)
def _nanmean_2_1(a: npt.NDArray[np.float32 | np.float64]) -> npt.NDArray[np.float64]:
    n = a.shape[0]
    output = np.empty(n, dtype=np.float64)
    for i in numba.prange(n):
        output[i] = np.nanmean(a[i, :])
    return output


@numba.njit(cache=True, parallel=True)
def _nanmean_3_0(a: npt.NDArray[np.float32 | np.float64]) -> npt.NDArray[np.float64]:
    _, n0, n1 = a.shape
    output = np.empty((n0, n1), dtype=np.float64)
    for i in numba.prange(n0):
        for j in range(n1):
            output[i, j] = np.nanmean(a[:, i, j])
    return output


@numba.njit(cache=True, parallel=True)
def _nanmean_3_1(a: npt.NDArray[np.float32 | np.float64]) -> npt.NDArray[np.float64]:
    n0, _, n1 = a.shape
    output = np.empty((n0, n1), dtype=np.float64)
    for i in numba.prange(n0):
        for j in range(n1):
            output[i, j] = np.nanmean(a[i, :, j])
    return output


@numba.njit(cache=True, parallel=True)
def _nanmean_3_2(a: npt.NDArray[np.float32 | np.float64]) -> npt.NDArray[np.float64]:
    n0, n1, _ = a.shape
    output = np.empty((n0, n1), dtype=np.float64)
    for i in numba.prange(n0):
        for j in range(n1):
            output[i, j] = np.nanmean(a[i, j, :])
    return output


@numba.njit(cache=True, parallel=True)
def _nanmean_3_01(a: npt.NDArray[np.float32 | np.float64]) -> npt.NDArray[np.float64]:
    n = a.shape[2]
    output = np.empty(n, dtype=np.float64)
    for i in numba.prange(n):
        output[i] = np.nanmean(a[:, :, i])
    return output


@numba.njit(cache=True, parallel=True)
def _nanmean_3_02(a: npt.NDArray[np.float32 | np.float64]) -> npt.NDArray[np.float64]:
    n = a.shape[1]
    output = np.empty(n, dtype=np.float64)
    for i in numba.prange(n):
        output[i] = np.nanmean(a[:, i, :])
    return output


@numba.njit(cache=True, parallel=True)
def _nanmean_3_12(a: npt.NDArray[np.float32 | np.float64]) -> npt.NDArray[np.float64]:
    n = a.shape[0]
    output = np.empty(n, dtype=np.float64)
    for i in numba.prange(n):
        output[i] = np.nanmean(a[i, :, :])
    return output


@numba.njit(cache=True, parallel=True)
def _nanmean_4_0(a: npt.NDArray[np.float32 | np.float64]) -> npt.NDArray[np.float64]:
    _, n0, n1, n2 = a.shape
    output = np.empty((n0, n1, n2), dtype=np.float64)
    for i in numba.prange(n0):
        for j in range(n1):
            for k in range(n2):
                output[i, j, k] = np.nanmean(a[:, i, j, k])
    return output


@numba.njit(cache=True, parallel=True)
def _nanmean_4_1(a: npt.NDArray[np.float32 | np.float64]) -> npt.NDArray[np.float64]:
    n0, _, n1, n2 = a.shape
    output = np.empty((n0, n1, n2), dtype=np.float64)
    for i in numba.prange(n0):
        for j in range(n1):
            for k in range(n2):
                output[i, j, k] = np.nanmean(a[i, :, j, k])
    return output


@numba.njit(cache=True, parallel=True)
def _nanmean_4_2(a: npt.NDArray[np.float32 | np.float64]) -> npt.NDArray[np.float64]:
    n0, n1, _, n2 = a.shape
    output = np.empty((n0, n1, n2), dtype=np.float64)
    for i in numba.prange(n0):
        for j in range(n1):
            for k in range(n2):
                output[i, j, k] = np.nanmean(a[i, j, :, k])
    return output


@numba.njit(cache=True, parallel=True)
def _nanmean_4_3(a: npt.NDArray[np.float32 | np.float64]) -> npt.NDArray[np.float64]:
    n0, n1, n2, _ = a.shape
    output = np.empty((n0, n1, n2), dtype=np.float64)
    for i in numba.prange(n0):
        for j in range(n1):
            for k in range(n2):
                output[i, j, k] = np.nanmean(a[i, j, k, :])
    return output


@numba.njit(cache=True, parallel=True)
def _nanmean_4_01(a: npt.NDArray[np.float32 | np.float64]) -> npt.NDArray[np.float64]:
    _, _, n0, n1 = a.shape
    output = np.empty((n0, n1), dtype=np.float64)
    for i in numba.prange(n0):
        for j in range(n1):
            output[i, j] = np.nanmean(a[:, :, i, j])
    return output


@numba.njit(cache=True, parallel=True)
def _nanmean_4_02(a: npt.NDArray[np.float32 | np.float64]) -> npt.NDArray[np.float64]:
    _, n0, _, n1 = a.shape
    output = np.empty((n0, n1), dtype=np.float64)
    for i in numba.prange(n0):
        for j in range(n1):
            output[i, j] = np.nanmean(a[:, i, :, j])
    return output


@numba.njit(cache=True, parallel=True)
def _nanmean_4_03(a: npt.NDArray[np.float32 | np.float64]) -> npt.NDArray[np.float64]:
    _, n0, n1, _ = a.shape
    output = np.empty((n0, n1), dtype=np.float64)
    for i in numba.prange(n0):
        for j in range(n1):
            output[i, j] = np.nanmean(a[:, i, j, :])
    return output


@numba.njit(cache=True, parallel=True)
def _nanmean_4_12(a: npt.NDArray[np.float32 | np.float64]) -> npt.NDArray[np.float64]:
    n0, _, _, n1 = a.shape
    output = np.empty((n0, n1), dtype=np.float64)
    for i in numba.prange(n0):
        for j in range(n1):
            output[i, j] = np.nanmean(a[i, :, :, j])
    return output


@numba.njit(cache=True, parallel=True)
def _nanmean_4_13(a: npt.NDArray[np.float32 | np.float64]) -> npt.NDArray[np.float64]:
    n0, _, n1, _ = a.shape
    output = np.empty((n0, n1), dtype=np.float64)
    for i in numba.prange(n0):
        for j in range(n1):
            output[i, j] = np.nanmean(a[i, :, j, :])
    return output


@numba.njit(cache=True, parallel=True)
def _nanmean_4_23(a: npt.NDArray[np.float32 | np.float64]) -> npt.NDArray[np.float64]:
    n0, n1, _, _ = a.shape
    output = np.empty((n0, n1), dtype=np.float64)
    for i in numba.prange(n0):
        for j in range(n1):
            output[i, j] = np.nanmean(a[i, j, :, :])
    return output


@numba.njit(cache=True, parallel=True)
def _nanmean_4_012(a: npt.NDArray[np.float32 | np.float64]) -> npt.NDArray[np.float64]:
    n = a.shape[3]
    output = np.empty(n, dtype=np.float64)
    for i in numba.prange(n):
        output[i] = np.nanmean(a[:, :, :, i])
    return output


@numba.njit(cache=True, parallel=True)
def _nanmean_4_013(a: npt.NDArray[np.float32 | np.float64]) -> npt.NDArray[np.float64]:
    n = a.shape[2]
    output = np.empty(n, dtype=np.float64)
    for i in numba.prange(n):
        output[i] = np.nanmean(a[:, :, i, :])
    return output


@numba.njit(cache=True, parallel=True)
def _nanmean_4_023(a: npt.NDArray[np.float32 | np.float64]) -> npt.NDArray[np.float64]:
    n = a.shape[1]
    output = np.empty(n, dtype=np.float64)
    for i in numba.prange(n):
        output[i] = np.nanmean(a[:, i, :, :])
    return output


@numba.njit(cache=True, parallel=True)
def _nanmean_4_123(a: npt.NDArray[np.float32 | np.float64]) -> npt.NDArray[np.float64]:
    n = a.shape[0]
    output = np.empty(n, dtype=np.float64)
    for i in numba.prange(n):
        output[i] = np.nanmean(a[i, :, :, :])
    return output


nanmean_funcs = {
    2: {
        0: _nanmean_2_0,
        1: _nanmean_2_1,
        frozenset({0}): _nanmean_2_0,
        frozenset({1}): _nanmean_2_1,
    },
    3: {
        0: _nanmean_3_0,
        1: _nanmean_3_1,
        2: _nanmean_3_2,
        frozenset({0}): _nanmean_3_0,
        frozenset({1}): _nanmean_3_1,
        frozenset({2}): _nanmean_3_2,
        frozenset({0, 1}): _nanmean_3_01,
        frozenset({0, 2}): _nanmean_3_02,
        frozenset({1, 2}): _nanmean_3_12,
    },
    4: {
        0: _nanmean_4_0,
        1: _nanmean_4_1,
        2: _nanmean_4_2,
        3: _nanmean_4_3,
        frozenset({0}): _nanmean_4_0,
        frozenset({1}): _nanmean_4_1,
        frozenset({2}): _nanmean_4_2,
        frozenset({3}): _nanmean_4_3,
        frozenset({0, 1}): _nanmean_4_01,
        frozenset({0, 2}): _nanmean_4_02,
        frozenset({0, 3}): _nanmean_4_03,
        frozenset({1, 2}): _nanmean_4_12,
        frozenset({1, 3}): _nanmean_4_13,
        frozenset({2, 3}): _nanmean_4_23,
        frozenset({0, 1, 2}): _nanmean_4_012,
        frozenset({0, 1, 3}): _nanmean_4_013,
        frozenset({0, 2, 3}): _nanmean_4_023,
        frozenset({1, 2, 3}): _nanmean_4_123,
    },
}


def fast_nanmean(
    a: npt.NDArray[np.float32 | np.float64], axis: int | Iterable[int] | None = None
) -> npt.NDArray[np.float32 | np.float64] | float:
    """A fast, parallelized arithmetic mean for floating point arrays that ignores NaNs.

    Parameters
    ----------
    a
        A numpy array of floats.
    axis
        Axis or iterable of axis along which the means are computed. If `None`, the mean
        of the flattend array is computed.

    Returns
    -------
    numpy.ndarray or float
        The calculated mean. The output array is always C-contiguous.

    Note
    ----
    Parallelization is only applied for :code:`N`-dimensional arrays with :code:`N <= 4`
    and :code:`len(axis) < N`. For bigger :code:`N`, :obj:`numbagg.nanmean` is used. For
    calculating the average of a flattened array (:code:`axis = None` or
    :code:`len(axis) == N`), the :obj:`numba` implemenation of :obj:`numpy.nanmean` is
    used. This function does not keep the input dimensions, i.e., the output is
    squeezed.

    """
    if a.ndim == 1 or axis is None:
        return _nanmean_all(a)
    elif a.ndim > 4:
        return np.ascontiguousarray(numbagg.nanmean(a, axis))
    if hasattr(axis, "__iter__"):
        if len(axis) == a.ndim:
            return _nanmean_all(a)
        axis = frozenset(x % a.ndim for x in axis)
    else:
        axis = axis % a.ndim
    return nanmean_funcs[a.ndim][axis](a).astype(a.dtype)


def _fast_nanmean_skipcheck(
    a: npt.NDArray[np.float32 | np.float64], axis: int | Iterable[int]
) -> npt.NDArray[np.float32 | np.float64] | float:
    """A version of `fast_nanmean` with near-zero overhead. Meant for internal use.

    Strict assumptions on the input parameters allow skipping some checks.

    Parameters
    ----------
    a
        A numpy array of floats. :code:`a.ndim` must be one of 2, 3, and 4.
    axis
        Axis or iterable of axis along which the means are computed. All elements must
        be nonnegative integers that are less than or equal to :code:`a.ndim`, i.e.,
        negative indexing is not allowed.

    Returns
    -------
    numpy.ndarray or float
        The calculated mean. The output array is always C-contiguous.

    """
    if hasattr(axis, "__iter__"):
        if len(axis) == a.ndim:
            return _nanmean_all(a)
        axis = frozenset(axis)
    return nanmean_funcs[a.ndim][axis](a).astype(a.dtype)


if __name__ == "__main__":
    for nd, funcs in nanmean_funcs.items():
        x = np.random.RandomState(42).randn(*((30,) * nd))
        for axis, func in funcs.items():
            if isinstance(axis, frozenset):
                axis = tuple(axis)
            if not np.allclose(np.nanmean(x, axis), fast_nanmean(x, axis)):
                print(func)



"""Class-based dynamic functions for fitting.

These functions are not limited to a single function form, and can be used to create
complex models.
"""

__all__ = [
    "get_args_kwargs",
    "DynamicFunction",
    "PolyFunc",
    "MultiPeakFunction",
    "FermiEdge2dFunc",
]
import functools
import inspect
from collections.abc import Callable

import numpy as np
import numpy.typing as npt
import xarray as xr

from erlab.analysis.fit.functions.general import (
    TINY,
    do_convolve,
    do_convolve_y,
    fermi_dirac,
    gaussian_wh,
    lorentzian_wh,
)
from erlab.constants import kb_eV


def get_args_kwargs(func) -> tuple[list[str], dict[str, object]]:
    """Get all argument names and default values from a function signature.

    Parameters
    ----------
    func
        The function to inspect.

    Returns
    -------
    args : list of str
        A list of argument names with no default value.
    args_default : dict
        A dictionary of keyword arguments with their default values.

    Note
    ----
    This function does not support function signatures containing varargs.

    Example
    -------

    >>> def my_func(a, b=10):
    ...     pass
    >>> get_args_kwargs(my_func)
    (['a'], {'b': 10})

    """
    args = []
    args_default = {}
    sig = inspect.signature(func)
    for fnam, fpar in sig.parameters.items():
        if fpar.kind == fpar.VAR_POSITIONAL or fpar.kind == fpar.VAR_KEYWORD:
            raise ValueError(f"varargs '*{fnam}' is not supported")
        elif fpar.default == fpar.empty:
            args.append(fnam)
        else:
            args_default[fnam] = fpar.default

    return args, args_default


class DynamicFunction:
    """Base class for dynamic functions.

    Dynamic functions exploits the way `lmfit` handles asteval functions in
    `lmfit.Model._parse_params`.
    """

    @property
    def __name__(self) -> str:
        return self.__class__.__name__

    @property
    def argnames(self) -> list[str]:
        return ["x"]

    @property
    def kwargs(self) -> dict[str, int | float]:
        return {}

    def __call__(self, x: npt.NDArray[np.float64], **params) -> npt.NDArray[np.float64]:
        raise NotImplementedError("Must be overloaded in child classes")


class PolyFunc(DynamicFunction):
    """A callable class for a arbitrary degree polynomial.

    Parameters
    ----------
    degree
        The degree of the polynomial.

    """

    def __init__(self, degree: int = 1) -> None:
        super().__init__()
        self.degree = degree

    @property
    def argnames(self) -> list[str]:
        return ["x"] + [f"c{i}" for i in range(self.degree + 1)]

    def __call__(self, x, *coeffs, **params):
        if len(coeffs) != self.degree + 1:
            coeffs = [params[f"c{d}"] for d in range(self.degree + 1)]
        if isinstance(x, np.ndarray):
            return np.polynomial.polynomial.polyval(x, coeffs)
        else:
            coeffs = xr.DataArray(coeffs, coords={"degree": np.arange(self.degree + 1)})
            return xr.polyval(x, coeffs)


class MultiPeakFunction(DynamicFunction):
    """A callable class for a multi-peak model.

    Parameters
    ----------
    npeaks
        The number of peaks to fit.
    peak_shapes
        The shape(s) of the peaks in the model. If a list of strings is provided, each
        string represents the shape of a peak. If a single string is provided, it will
        be split by spaces to create a list of peak shapes. If not provided, the default
        peak shape will be used for all peaks.
    fd
        Flag indicating whether the model should be multiplied by the Fermi-Dirac
        distribution. This adds three parameters to the model: `efermi`, `temp`, and
        `offset`, each corresponding to the Fermi level, temperature in K, and constant
        background.
    convolve
        Flag indicating whether the model should be convolved with a gaussian kernel. If
        `True`, adds a `resolution` parameter to the model, corresponding to the FWHM of
        the gaussian kernel.

    """

    PEAK_SHAPES: dict[Callable, list[str]] = {
        lorentzian_wh: ["lorentzian", "lor", "l"],
        gaussian_wh: ["gaussian", "gauss", "g"],
    }

    DEFAULT_PEAK: str = "lorentzian"

    def __init__(
        self,
        npeaks: int,
        peak_shapes: list[str] | str | None = None,
        fd: bool = True,
        convolve: bool = True,
    ):
        super().__init__()
        self.npeaks = npeaks
        self.fd = fd
        self.convolve = convolve

        if peak_shapes is None:
            peak_shapes = [self.DEFAULT_PEAK] * self.npeaks
        if isinstance(peak_shapes, str):
            peak_shapes = peak_shapes.split(" ")

        if len(peak_shapes) == 1:
            peak_shapes = peak_shapes * self.npeaks
        elif len(peak_shapes) != self.npeaks:
            raise ValueError("Number of peaks does not match given peak shapes")

        self._peak_shapes = peak_shapes

        self._peak_funcs = [None] * self.npeaks
        for i, name in enumerate(self._peak_shapes):
            for fcn, aliases in self.PEAK_SHAPES.items():
                if name in aliases:
                    self._peak_funcs[i] = fcn

        if None in self._peak_funcs:
            raise ValueError("Invalid peak name")

    @functools.cached_property
    def peak_all_args(self) -> dict[Callable, dict[str, list | dict]]:
        res = {}
        for func in self.PEAK_SHAPES:
            res[func] = dict(zip(("args", "kwargs"), get_args_kwargs(func)))
        return res

    @functools.cached_property
    def peak_argnames(self) -> dict[Callable, list[str]]:
        res = {}
        for func in self.PEAK_SHAPES:
            res[func] = self.peak_all_args[func]["args"][1:] + list(
                self.peak_all_args[func]["kwargs"].keys()
            )
        return res

    @property
    def peak_funcs(self) -> list[Callable]:
        return self._peak_funcs

    @property
    def argnames(self) -> list[str]:
        args = ["x"]
        for i, func in enumerate(self.peak_funcs):
            args += [f"p{i}_{arg}" for arg in self.peak_all_args[func]["args"][1:]]
        return args

    @property
    def kwargs(self):
        kws = [
            ("lin_bkg", 0),
            ("const_bkg", 0),
        ]
        if self.fd:
            kws += [
                ("efermi", 0.0),  # fermi level
                ("temp", 30),  # temperature
                ("offset", 0),
            ]
        if self.convolve:
            kws += [("resolution", 0.02)]

        for i, func in enumerate(self.peak_funcs):
            for arg, val in self.peak_all_args[func]["kwargs"].items():
                kws.append((f"p{i}_{arg}", val))
        return kws

    def eval_peak(self, index: int, x: npt.NDArray[np.float64], **params: dict):
        return self.peak_funcs[index](
            x, **{k[3:]: v for k, v in params.items() if k.startswith(f"p{index}")}
        )

    def pre_call(
        self, x: npt.NDArray[np.float64], **params: dict
    ) -> npt.NDArray[np.float64]:
        x = np.asarray(x).copy()
        y = np.zeros_like(x)

        for i, func in enumerate(self.peak_funcs):
            y += func(
                x, **{arg: params[f"p{i}_{arg}"] for arg in self.peak_argnames[func]}
            )

        y += params["lin_bkg"] * x + params["const_bkg"]

        if self.fd:
            y *= fermi_dirac(x, center=params["efermi"], temp=params["temp"])
            y += params["offset"]

        return y

    def __call__(
        self, x: npt.NDArray[np.float64], **params: dict
    ) -> npt.NDArray[np.float64]:
        if self.convolve:
            return do_convolve(x, self.pre_call, **params)
        else:
            return self.pre_call(x, **params)


class FermiEdge2dFunc(DynamicFunction):
    def __init__(self, degree=1) -> None:
        super().__init__()
        self.poly = PolyFunc(degree)

    @property
    def argnames(self) -> list[str]:
        return ["eV", "alpha"] + self.poly.argnames[1:]

    @property
    def kwargs(self):
        return [
            ("temp", 30.0),
            ("lin_bkg", 0.0),
            ("const_bkg", 1.0),
            ("offset", 0.0),
            ("resolution", 0.02),
        ]

    def pre_call(self, eV, alpha, **params: dict):
        center = self.poly(
            np.asarray(alpha),
            *[params.pop(f"c{i}") for i in range(self.poly.degree + 1)],
        )
        return (params["const_bkg"] - params["offset"] + params["lin_bkg"] * eV) / (
            1 + np.exp((1.0 * eV - center) / max(TINY, params["temp"] * kb_eV))
        ) + params["offset"]

    def __call__(self, eV, alpha, **params: dict):
        return do_convolve_y(eV, alpha, self.pre_call, **params).flatten()



"""Utilities for curve fitting.

.. currentmodule:: erlab.analysis.fit

Modules
=======

.. autosummary::
   :toctree:

   functions
   models
   spline
   minuit

"""

__all__ = [
    "ExtendedAffineBroadenedFD",
    "PolynomialModel",
    "MultiPeakModel",
    "LeastSq",
    "Minuit",
]

from erlab.analysis.fit.minuit import LeastSq, Minuit
from erlab.analysis.fit.models import (
    ExtendedAffineBroadenedFD,
    MultiPeakModel,
    PolynomialModel,
)



import sys
import numpy as np
import pyqtgraph as pg

from arpes.analysis.derivative import curvature, minimum_gradient
from erlab.interactive.utilities import parse_data, AnalysisWidgetBase

from qtpy import QtCore, QtWidgets
from scipy.ndimage import gaussian_filter, uniform_filter

# def move_mean_centered(a, window, min_count=None, axis=-1):
#     w = (window - 1) // 2
#     shift = w + 1
#     if min_count is None:
#         min_count = w + 1
#     pad_width = [(0, 0)] * a.ndim
#     pad_width[axis] = (0, shift)
#     a = np.pad(a, pad_width, constant_values=np.nan)
#     val = bn.move_mean(a, window, min_count=min_count, axis=axis)
#     return val[(slice(None),) * (axis % a.ndim) + (slice(w, -1),)]


# def move_mean_centered_multiaxis(a, window_list, min_count_list=None):
#     w_list = [(window - 1) // 2 for window in window_list]
#     pad_width = [(0, 0)] * a.ndim
#     slicer = [
#         slice(None),
#     ] * a.ndim
#     if min_count_list is None:
#         min_count_list = [w + 1 for w in w_list]
#     for axis in range(a.ndim):
#         pad_width[axis] = (0, w_list[axis] + 1)
#         slicer[axis] = slice(w_list[axis], -1)
#     a = np.pad(a, pad_width, constant_values=np.nan)
#     val = move_mean(
#         a,
#         numba.typed.List(window_list),
#         numba.typed.List(min_count_list),
#     )
#     return val[tuple(slicer)]


# def move_mean(a, window, min_count):
#     if a.ndim == 3:
#         return move_mean3d(a, window, min_count)
#     elif a.ndim == 2:
#         return move_mean2d(a, window, min_count)
#     else:
#         raise NotImplementedError


# pg.setConfigOption('useNumba', True)

__all__ = ["dtool"]


# def boxcarfilt2(array, window_list, min_count_list=None):
#     return move_mean_centered_multiaxis(array, window_list, min_count_list)


def laplacian_O3(f, *varargs):
    """
    Third order accurate, 5-point formula. Not really laplacian, but second derivative
    along each axis. Sum the outputs to get laplacian.
    """
    N = len(f.shape)  # number of dimensions
    n = len(varargs)
    if n == 0:
        dx = [1.0] * N
    elif n == 1:
        dx = [varargs[0]] * N
    elif n == N:
        dx = list(varargs)
    else:
        raise SyntaxError("invalid number of arguments")

    # use central differences on interior and first differences on endpoints

    # print dx
    outvals = []

    # create slice objects --- initially all are [:, :, ..., :]
    slice0 = [slice(None)] * N
    slice1 = [slice(None)] * N
    slice2 = [slice(None)] * N
    slice3 = [slice(None)] * N
    slice4 = [slice(None)] * N

    otype = f.dtype.char
    if otype not in ["f", "d", "F", "D"]:
        otype = "d"

    for axis in range(N):
        # select out appropriate parts for this dimension
        out = np.zeros(f.shape, f.dtype.char)

        # http://www.sitmo.com/eq/262 (3rd order accurate)
        slice0[axis] = slice(2, -2)
        slice1[axis] = slice(None, -4)
        slice2[axis] = slice(1, -3)
        slice3[axis] = slice(3, -1)
        slice4[axis] = slice(4, None)
        # 1D equivalent -- out[2:-2] = (-f[:-4] + 16*f[1:-3] + -30*f[2:-2] +
        # 16*f[3:-1] - f[4:])/12.0
        out[slice0] = (
            -f[slice1]
            + 16.0 * f[slice2]
            - 30.0 * f[slice0]
            + 16.0 * f[slice3]
            - f[slice4]
        ) / 12.0

        # http://www.sitmo.com/eq/260 (2nd order accurate; there's also a 3rd
        # order accurate that requires 5 points)
        slice0[axis] = slice(None, 2)
        slice1[axis] = slice(3, 5)
        slice2[axis] = slice(2, 4)
        slice3[axis] = slice(1, 3)
        # 1D equivalent -- out[0:2] = 2*f[0:2] - 5*f[1:3] + 4*f[2:4] - f[3:5]
        out[slice0] = 2.0 * f[slice0] - 5.0 * f[slice3] + 4.0 * f[slice2] - f[slice1]

        slice0[axis] = slice(-2, None)
        slice1[axis] = slice(-5, -3)
        slice2[axis] = slice(-4, -2)
        slice3[axis] = slice(-3, -1)
        # 1D equivalent -- out[0:2] = 2*f[0:2] - 5*f[1:3] + 4*f[2:4] - f[3:5]
        out[slice0] = 2.0 * f[slice0] - 5.0 * f[slice3] + 4.0 * f[slice2] - f[slice1]

        # divide by step size
        axis_dx = dx[axis]
        outvals.append(out / (axis_dx * axis_dx))

        # reset the slice object in this dimension to ":"
        slice0[axis] = slice(None)
        slice1[axis] = slice(None)
        slice2[axis] = slice(None)
        slice3[axis] = slice(None)
        slice4[axis] = slice(None)

    if N == 1:
        return outvals[0]
    else:
        return outvals


class NoiseToolNew(AnalysisWidgetBase):
    def __init__(self, data, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_input(data)

    # def get_smoothfunc(self, amount, num, method="boxcar"):
    #     if method == "boxcar":
    #         def func(data):
    #             out = data.copy(deep=True)
    #             vals = data.values
    #             for _ in range(num):
    #                 vals = uniform_filter(
    #                     vals,
    #                     amount,
    #                     mode="constant",
    #                     origin=(0, 0),
    #                 )
    #             out.values = vals
    #             return out
    #     elif method == "gaussian":
    #         def func(data):
    #             out = data.copy(deep=True)
    #             vals = data.values
    #             sigma = [(amount[0] - 1) / 3, (amount[1] - 1) / 3]
    #             for _ in range(num):
    #                 vals = gaussian_filter(vals, sigma=sigma)
    #             out.values = vals
    #             return out
    #     return func


class DerivativeTool(QtWidgets.QMainWindow):
    def __init__(self, data=None, *args, **kwargs):
        super().__init__()

        self._main = QtWidgets.QWidget()
        self.setCentralWidget(self._main)
        self.layout = QtWidgets.QGridLayout(self._main)
        plotwidget = pg.GraphicsLayoutWidget(show=True)

        self.hists = [
            pg.HistogramLUTItem(),
            pg.HistogramLUTItem(),
        ]
        self.ax1 = plotwidget.addPlot(0, 0, 1, 1)
        plotwidget.addItem(self.hists[0])
        plotwidget.nextRow()
        self.ax2 = plotwidget.addPlot(1, 0, 1, 1)
        plotwidget.addItem(self.hists[1])
        self.axes = [self.ax1, self.ax2]
        self.ax2.setYLink(self.ax1)
        self.ax2.setXLink(self.ax1)

        self.images = [
            pg.ImageItem(axisOrder="row-major"),
            pg.ImageItem(axisOrder="row-major"),
        ]
        for i, img in enumerate(self.images):
            self.axes[i].addItem(img)
            self.hists[i].setImageItem(img)
            # img.sigImageChanged.connect()

        self.use_gaussfilt = False

        self.win_list = [1, 1]
        self.smooth_num = 1
        self.smooth2_num = 1

        self._smooth_group = QtWidgets.QGroupBox("Smoothing")
        smooth_layout = QtWidgets.QGridLayout(self._smooth_group)
        self._smooth_mode_check = QtWidgets.QCheckBox(self._smooth_group)
        self._smooth_mode_check.setChecked(self.use_gaussfilt)
        smooth_mode_label = QtWidgets.QLabel("Gaussian filter")
        smooth_mode_label.setBuddy(self._smooth_mode_check)
        self._smooth_x_spin = QtWidgets.QSpinBox(self._smooth_group)
        self._smooth_y_spin = QtWidgets.QSpinBox(self._smooth_group)
        self._smooth_n_spin = QtWidgets.QSpinBox(self._smooth_group)
        self._smooth_n_spin.setRange(1, 100)
        smooth_x_spin_label = QtWidgets.QLabel("x")
        smooth_y_spin_label = QtWidgets.QLabel("y")
        smooth_n_spin_label = QtWidgets.QLabel("n")
        smooth_layout.addWidget(smooth_mode_label, 0, 0)
        smooth_layout.addWidget(self._smooth_mode_check, 0, 1)
        smooth_layout.addWidget(smooth_x_spin_label, 1, 0)
        smooth_layout.addWidget(self._smooth_x_spin, 1, 1)
        smooth_layout.addWidget(smooth_y_spin_label, 2, 0)
        smooth_layout.addWidget(self._smooth_y_spin, 2, 1)
        smooth_layout.addWidget(smooth_n_spin_label, 3, 0)
        smooth_layout.addWidget(self._smooth_n_spin, 3, 1)

        self._color_group = QtWidgets.QGroupBox("Color")
        color_layout = QtWidgets.QGridLayout(self._color_group)
        self._color_range_spin = (
            QtWidgets.QDoubleSpinBox(self._smooth_group),
            QtWidgets.QDoubleSpinBox(self._smooth_group),
        )
        for i, s in enumerate(self._color_range_spin):
            s.setRange(0, 100)
            s.setValue(10)
            s.setSingleStep(0.1)
            s.setSuffix(" %")
            s.setDecimals(1)
            color_layout.addWidget(s, i, 0)
        self._color_range_spin[1].setValue(1)

        self._smooth2_group = QtWidgets.QGroupBox("2nd Smoothing")
        smooth2_layout = QtWidgets.QGridLayout(self._smooth2_group)
        self._smooth2_x_spin = QtWidgets.QSpinBox(self._smooth2_group)
        self._smooth2_y_spin = QtWidgets.QSpinBox(self._smooth2_group)
        self._smooth2_n_spin = QtWidgets.QSpinBox(self._smooth2_group)
        self._smooth2_x_spin_label = QtWidgets.QLabel("x")
        self._smooth2_y_spin_label = QtWidgets.QLabel("y")
        self._smooth2_n_spin_label = QtWidgets.QLabel("n")
        self._smooth2_n_spin.setRange(1, 100)
        smooth2_layout.addWidget(self._smooth2_x_spin_label, 1, 0)
        smooth2_layout.addWidget(self._smooth2_x_spin, 1, 1)
        smooth2_layout.addWidget(self._smooth2_y_spin_label, 2, 0)
        smooth2_layout.addWidget(self._smooth2_y_spin, 2, 1)
        smooth2_layout.addWidget(self._smooth2_n_spin_label, 3, 0)
        smooth2_layout.addWidget(self._smooth2_n_spin, 3, 1)

        self._deriv_group = QtWidgets.QGroupBox("Derivative")
        deriv_layout = QtWidgets.QGridLayout(self._deriv_group)
        self._deriv_axis_combo = QtWidgets.QComboBox(self._deriv_group)
        self._deriv_axis_combo.addItems(["x", "y"])
        deriv_layout.addWidget(self._deriv_axis_combo, 0, 0)

        self._curv_group = QtWidgets.QGroupBox("Curvature")
        curv_layout = QtWidgets.QGridLayout(self._curv_group)
        self._curv_alpha_spin = QtWidgets.QDoubleSpinBox(self._curv_group)
        self._curv_alpha_spin.setRange(-30, 30)
        self._curv_alpha_spin.setValue(0)
        self._curv_alpha_spin.setSingleStep(0.05)
        curv_layout.addWidget(self._curv_alpha_spin, 0, 0)

        self._mingrad_group = QtWidgets.QGroupBox("Minimum Gradient")
        mingrad_layout = QtWidgets.QGridLayout(self._mingrad_group)
        self._mingrad_delta_spin = QtWidgets.QSpinBox(self._mingrad_group)
        self._mingrad_delta_spin.setRange(1, 100)
        self._mingrad_delta_spin.setValue(1)
        mingrad_layout.addWidget(self._mingrad_delta_spin, 0, 0)

        self._deriv_tab = QtWidgets.QWidget()
        deriv_content = QtWidgets.QGridLayout(self._deriv_tab)
        deriv_content.addWidget(self._smooth2_group, 0, 0)
        deriv_content.addWidget(self._deriv_group, 0, 1)

        self._curv_tab = QtWidgets.QWidget()
        curv_content = QtWidgets.QGridLayout(self._curv_tab)
        curv_content.addWidget(self._curv_group, 0, 0)

        self._mingrad_tab = QtWidgets.QWidget()
        mingrad_content = QtWidgets.QGridLayout(self._mingrad_tab)
        mingrad_content.addWidget(self._mingrad_group, 0, 0)

        self.tabwidget = QtWidgets.QTabWidget()
        self.tabwidget.addTab(self._deriv_tab, "Derivative")
        self.tabwidget.addTab(self._curv_tab, "Curvature")
        self.tabwidget.addTab(self._mingrad_tab, "Minimum Gradient")
        self.tabwidget.setSizePolicy(
            QtWidgets.QSizePolicy.MinimumExpanding,
            QtWidgets.QSizePolicy.Maximum,
        )

        self.layout.addWidget(self._smooth_group, 0, 0, 1, 1)
        self.layout.addWidget(self._color_group, 0, 1, 1, 1)
        self.layout.addWidget(self.tabwidget, 0, 2, 1, 1)
        self.layout.addWidget(plotwidget, 1, 0, 2, 3)

        self.use_curv = False
        self.use_mingrad = False
        self.use_deriv = True

        if data is not None:
            self.set_data(data)

        self._smooth_mode_check.stateChanged.connect(self.set_gauss_smooth)
        self._smooth_x_spin.valueChanged.connect(
            lambda val, ax=1: self.smooth_data(ax, val)
        )
        self._smooth_y_spin.valueChanged.connect(
            lambda val, ax=0: self.smooth_data(ax, val)
        )
        self._smooth_n_spin.valueChanged.connect(self.set_smooth_num)
        self._smooth2_x_spin.valueChanged.connect(lambda: self.deriv_data())
        self._smooth2_y_spin.valueChanged.connect(lambda: self.deriv_data())
        self._smooth2_n_spin.valueChanged.connect(self.set_smooth2_num)
        self._deriv_axis_combo.currentTextChanged.connect(self.deriv_data)
        self._curv_alpha_spin.valueChanged.connect(
            lambda beta: self.curv_data(beta=beta)
        )
        self._mingrad_delta_spin.valueChanged.connect(
            lambda delta: self.mingrad_data(delta=delta)
        )
        for s in self._color_range_spin:
            s.valueChanged.connect(self._result_update_lims)

    def get_cutoff(self, data, cutoff=None):
        q3, q1 = np.percentile(data, [75, 25])
        ql = q1 - 1.5 * (q3 - q1)
        qu = q3 + 1.5 * (q3 - q1)
        i_qu = 100 * (data > qu).mean()
        i_ql = 100 * (data < ql).mean()
        for s in self._color_range_spin:
            s.blockSignals(True)
        self._color_range_spin[0].setMaximum(i_qu)
        self._color_range_spin[1].setMaximum(i_ql)
        for s in self._color_range_spin:
            s.blockSignals(False)
        if cutoff is None:
            # cutoff = [30, 30]
            cutoff = [s.value() for s in self._color_range_spin]
        else:
            try:
                cutoff = list(cutoff.__iter__)
            except AttributeError:
                cutoff = [cutoff] * 2

        pu, pl = np.percentile(data, [100 - cutoff[0], cutoff[1]])
        mn = max(min(pl, ql), data.min())
        mx = min(max(pu, qu), data.max())
        return [mn, mx]

    def set_gauss_smooth(self, value):
        self.use_gaussfilt = value
        self.smooth_data()

    def _result_update_lims(self):
        self.images[1].setImage(
            self.data_s_d.values,
            rect=self._lims_to_rect(1, 0),
            levels=self.get_cutoff(self.data_s_d.values),
        )

    def set_smooth_num(self, n):
        self.smooth_num = n
        self.smooth_data()

    def set_smooth2_num(self, n):
        self.smooth2_num = n
        self.deriv_data()

    def mingrad_data(self, delta=1):
        self.use_curv = False
        self.use_mingrad = True
        self.use_deriv = False
        # self.data_s_d = self.data_s.copy(deep=True)
        self.data_s_d = -minimum_gradient(self.data_s, delta=delta)
        self._result_update_lims()

    def curv_data(self, alpha=1, beta=None):
        self.use_curv = True
        self.use_deriv = False
        self.use_mingrad = False
        if beta is None:
            beta = self._curv_alpha_spin.value()
        self.data_s_d = self.data_s.copy(deep=True)
        self.data_s_d.values = curvature(
            self.data_s / self.data_s.max(), alpha=alpha, beta=beta, values=True
        )
        # m = self.data_s_d.max()
        # self.data_s_d /= m
        self._result_update_lims()

    def deriv_data(self, axis=None):
        self.use_curv = False
        self.use_mingrad = False
        self.use_deriv = True
        if axis is None:
            axis = self._deriv_axis_combo.currentText()
        if axis == "x":
            axis = 0
        elif axis == "y":
            axis = 1
        # axis = self.data_dims[axis]
        self.data_s_d = self.data_s.copy(deep=True)
        self.data_s_d.values = np.gradient(self.data_s.values, axis=axis)
        amount = [self._smooth2_y_spin.value(), self._smooth2_x_spin.value()]
        self.use_gaussfilt = self._smooth_mode_check.isChecked()
        self.data_s_d.values = self.smoothfunc(
            self.data_s_d.values, amount, self.smooth2_num
        )
        self.data_s_d.values = np.gradient(self.data_s_d.values, axis=axis)
        # self.data_s_d = d1_along_axis(self.data_s_d, axis=axis)
        self._result_update_lims()

    def _get_middle_index(self, x):
        return len(x) // 2 - (1 if len(x) % 2 == 0 else 0)

    def smoothfunc(self, data, amount, num):
        if not self.use_gaussfilt:
            for _ in range(num):
                data = uniform_filter(data, amount, mode="constant", origin=(0, 0))
        else:
            sigma = [(amount[0] - 1) / 3, (amount[1] - 1) / 3]
            for _ in range(num):
                data = gaussian_filter(data, sigma=sigma, truncate=5)
        return data

    def smooth_data(self, axis=None, amount=None):
        if axis is not None:
            self.win_list[axis] = amount
        else:
            self.win_list = [self._smooth_y_spin.value(), self._smooth_x_spin.value()]
        self.use_gaussfilt = self._smooth_mode_check.isChecked()

        self.data_s = self.data.copy(deep=True)
        vals = self.data_s.values

        self.data_s.values = self.smoothfunc(vals, self.win_list, self.smooth_num)

        self.images[0].setImage(self.data_s.values, rect=self._lims_to_rect(1, 0))
        if self.use_deriv:
            self.deriv_data()
        elif self.use_curv:
            self.curv_data()
        elif self.use_mingrad:
            self.mingrad_data()

    def _lims_to_rect(self, i, j):
        x = self.data_lims[i][0] - self.data_incs[i]
        y = self.data_lims[j][0] - self.data_incs[j]
        w = self.data_lims[i][-1] - x
        h = self.data_lims[j][-1] - y
        x += 0.5 * self.data_incs[i]
        y += 0.5 * self.data_incs[j]
        return QtCore.QRectF(x, y, w, h)

    def set_data(self, data):
        self.data = parse_data(data).fillna(0)
        self.data_s = self.data.copy(deep=True)
        self.vals = self.data.values
        self.data_dims = self.data.dims
        self.data_coords = tuple(self.data[dim].values for dim in self.data_dims)
        self.data_incs = tuple(c[1] - c[0] for c in self.data_coords)
        self.data_lims = tuple((c[0], c[-1]) for c in self.data_coords)
        self.data_shape = self.data.shape
        self.data_origin = tuple(self._get_middle_index(c) for c in self.data_coords)
        self._smooth_x_spin.setRange(1, self.data_shape[0] - 1)
        self._smooth_y_spin.setRange(1, self.data_shape[1] - 1)
        self._smooth2_x_spin.setRange(1, self.data_shape[0] - 1)
        self._smooth2_y_spin.setRange(1, self.data_shape[1] - 1)
        self.smooth_data()
        # self.images[0].setImage(self.vals,
        # rect=self._lims_to_rect(1, 0))


def dtool(data, *args, **kwargs):
    qapp = QtWidgets.QApplication.instance()
    if not qapp:
        qapp = QtWidgets.QApplication(sys.argv)
    qapp.setApplicationName("Dispersive features analysis")
    app = DerivativeTool(data, *args, **kwargs)
    qapp.setStyle("Fusion")
    app.show()
    app.activateWindow()
    app.raise_()
    qapp.exec()


if __name__ == "__main__":
    import erlab.io

    gkmk_cvs = erlab.io.load_als_bl4(
        "/Users/khan/Documents/ERLab/CsV3Sb5/2021_Dec_ALS_CV3Sb5/211217 ALS BL4/csvtisb1/f_003.pxt"
    )
    alpha_new = np.linspace(gkmk_cvs.alpha[0], gkmk_cvs.alpha[-1], 1000)
    eV_new = np.linspace(gkmk_cvs.eV[0], gkmk_cvs.eV[-1], 2000)
    gkmk_cvs = gkmk_cvs.interp(alpha=alpha_new, eV=eV_new)
    gkmk_cvs = gkmk_cvs.sel(
        alpha=slice(*np.rad2deg((-0.25, 0.25))), eV=slice(-1.25, 0.15)
    )
    dtool(gkmk_cvs, bench=False)



"""Various helper functions and extensions to pyqtgraph."""

from __future__ import annotations

import re
import sys
import warnings
from typing import Any, Literal

import numpy as np
import numpy.typing as npt
import pyperclip
import pyqtgraph as pg
import xarray as xr
from qtpy import QtCore, QtGui, QtWidgets
from superqt import QDoubleSlider

from erlab.interactive.colors import pg_colormap_powernorm

__all__ = [
    "parse_data",
    "copy_to_clipboard",
    "gen_single_function_code",
    "gen_function_code",
    "BetterSpinBox",
    "BetterAxisItem",
    "xImageItem",
    "ParameterGroup",
    "AnalysisWidgetBase",
    "AnalysisWindow",
    "DictMenuBar",
]


def parse_data(data):
    if isinstance(data, xr.Dataset):
        raise TypeError(
            "input argument data must be a xarray.DataArray or a "
            "numpy.ndarray. Create an xarray.DataArray "
            "first, either with indexing on the Dataset or by "
            "invoking the `to_array()` method."
        ) from None
    elif isinstance(data, np.ndarray):
        data = xr.DataArray(data)
    return data  # .astype(float, order="C")


def array_rect(data):
    data_coords = tuple(data[dim].values for dim in data.dims)
    data_incs = tuple(coord[1] - coord[0] for coord in data_coords)
    data_lims = tuple((coord[0], coord[-1]) for coord in data_coords)
    y, x = data_lims[0][0] - data_incs[0], data_lims[1][0] - data_incs[1]
    h, w = data_lims[0][-1] - y, data_lims[1][-1] - x
    y += 0.5 * data_incs[0]
    x += 0.5 * data_incs[1]
    return QtCore.QRectF(x, y, w, h)


def copy_to_clipboard(content: str | list[str]) -> str:
    """Convenience function for clipboard handling.

    Parameters
    ----------
    content
        The content to be copied.

    Returns
    -------
    str
        The copied content.

    """
    if isinstance(content, list):
        content = "\n".join(content)
    pyperclip.copy(content)
    return content


def process_arg(arg):
    if isinstance(arg, str):
        if arg.startswith("|") and arg.endswith("|"):
            arg = arg[1:-1]
        else:
            arg = f'"{arg}"'
    return arg


def gen_single_function_code(funcname: str, *args: tuple, **kwargs: dict):
    """Generates the string for a Python function call.

    The first argument is the name of the function, and subsequent arguments are
    passed as positional arguments. Keyword arguments are also supported.

    Parameters
    ----------
    funcname
        Name of the function.
    *args
        Mandatory arguments passed onto the function.
    **kwargs
        Keyword arguments passed onto the function.

    Returns
    -------
    code : str
        generated code.

    """
    if len(args) == 0 and len(kwargs) == 0:
        return f"{funcname}()"
    tab = "    "
    code = f"{funcname}(\n"
    for v in args:
        if isinstance(v, str):
            if v.startswith("|") and v.endswith("|"):
                v = v[1:-1]
            else:
                v = f'"{v}"'
        code += f"{tab}{v},\n"
    for k, v in kwargs.items():
        if isinstance(v, str):
            v = f'"{v}"'
        code += f"{tab}{k}={v},\n"
    code += ")"
    return code


def gen_function_code(copy: bool = True, **kwargs: dict):
    r"""Copies the Python code for function calls to the clipboard.

    The result can be copied to your clipboard in a form that can be pasted into an
    interactive Python session or Jupyter notebook cell.

    Parameters
    ----------
    copy
        If `True`, the code string is copied.
    **kwargs
        Dictionary where the keys are the string of the function call and the values are
        a list of function arguments. The last item, if a dictionary, is interpreted as
        keyword arguments.

    """
    code_list = []
    for k, v in kwargs.items():
        if not isinstance(v[-1], dict):
            v.append({})
        code_list.append(gen_single_function_code(k, *v[:-1], **v[-1]))

    code_str = "\n".join(code_list)

    if copy:
        return copy_to_clipboard(code_str)
    else:
        return code_str


class BetterSpinBox(QtWidgets.QAbstractSpinBox):
    """An improved spinbox.

    Signals
    ----------
    valueChanged
        Emitted when the value is changed.
    textChanged
        Emitted when the text is changed.

    Parameters
    ----------
    integer
        If `True`, the spinbox will only display integer values.
    compact
        Whether to reduce the height of the spinbox.
    discrete
        If `True` the spinbox will only step to pre-determined discrete values.

        If `False`, the spinbox will just add or subtract the predetermined
        increment when increasing or decreasing the step.
    decimals
        The precision of the spinbox. See the `significant` argument for the
        meaning. When `integer` is `True`, this argument is ignored.
    significant
        If `True`, `decimals` will specify the total number of significant digits,
        before or after the decimal point, ignoring leading zeros.

        If `False`, `decimals` will specify the total number of digits after the
        decimal point, including leading zeros.

        When `integer` or `scientific` is `True`, this argument is ignored.
    scientific
        Whether to print in scientific notation.
    value
        Initial value of the spinbox.

    """

    valueChanged = QtCore.Signal(object)  #: :meta private:
    textChanged = QtCore.Signal(object)  #: :meta private:

    def __init__(
        self,
        *args,
        integer: bool = False,
        compact: bool = True,
        discrete: bool = False,
        decimals: int = 3,
        significant: bool = False,
        scientific: bool = False,
        value: float = 0.0,
        **kwargs: dict,
    ):
        self._only_int = integer
        self._is_compact = compact
        self._is_discrete = discrete
        self._is_scientific = scientific
        self._decimal_significant = significant
        self.setDecimals(decimals)

        self._value = value
        self._lastvalue = None
        self._min = -np.inf
        self._max = np.inf
        self._step = 1 if self._only_int else 0.01

        kwargs.setdefault("correctionMode", self.CorrectionMode.CorrectToPreviousValue)
        kwargs.setdefault("keyboardTracking", False)

        # PyQt6 compatibility: set options with keyword arguments
        set_dict = {}
        for k in ["singleStep", "minimum", "maximum"]:
            set_dict[k] = kwargs.pop(k, None)
        super().__init__(*args, **kwargs)
        for k, v in set_dict.items():
            if v is not None:
                getattr(self, f"set{k[0].capitalize()}{k[1:]}")(v)

        self.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.MinimumExpanding,
            QtWidgets.QSizePolicy.Policy.Fixed,
        )
        self.editingFinished.connect(self.editingFinishedEvent)
        self._updateHeight()
        self._updateWidth()

        if self.isReadOnly():
            self.lineEdit().setReadOnly(True)
            self.setButtonSymbols(self.ButtonSymbols.NoButtons)
        self.setValue(self.value())

    def setDecimals(self, decimals):
        self._decimals = decimals

    def decimals(self):
        return self._decimals

    def setRange(self, mn, mx):
        self.setMinimum(min(mn, mx))
        self.setMaximum(max(mn, mx))

    def widthFromText(self, text):
        return QtGui.QFontMetrics(self.font()).boundingRect(text).width()

    def widthFromValue(self, value):
        return self.widthFromText(self.textFromValue(value))

    def setMaximum(self, mx):
        if self._only_int and np.isfinite(mx):
            mx = round(mx)
        elif np.isnan(mx):
            mx = np.inf
        self._max = mx
        self._updateWidth()

    def setMinimum(self, mn):
        if self._only_int and np.isfinite(mn):
            mn = round(mn)
        elif np.isnan(mn):
            mn = -np.inf
        self._min = mn
        self._updateWidth()

    def setSingleStep(self, step):
        if self._only_int:
            step = round(step)
        self._step = abs(step)

    def singleStep(self):
        return self._step

    def maximum(self):
        return self._max

    def minimum(self):
        return self._min

    def value(self):
        if self._only_int:
            return int(self._value)
        else:
            return self._value

    def text(self):
        return self.textFromValue(self.value())

    def textFromValue(self, value):
        if (not self._only_int) or (not np.isfinite(value)):
            if self._is_scientific:
                return np.format_float_scientific(
                    value,
                    precision=self.decimals(),
                    unique=False,
                    trim="k",
                    exp_digits=1,
                )
            else:
                return np.format_float_positional(
                    value,
                    precision=self.decimals(),
                    unique=False,
                    fractional=not self._decimal_significant,
                    trim="k",
                )
        else:
            return str(int(value))

    def valueFromText(self, text):
        if text == "":
            return np.nan
        if self._only_int:
            return int(text)
        else:
            return float(text)

    def stepBy(self, steps):
        inc = self.singleStep()
        if (
            all(np.isfinite([self.maximum(), self.minimum(), self.value()]))
            and self._is_discrete
        ):
            self.setValue(
                self.minimum()
                + inc
                * max(
                    min(
                        round((self.value() + steps * inc - self.minimum()) / inc),
                        int((self.maximum() - self.minimum()) / inc),
                    ),
                    0,
                )
            )
        elif steps > 0:
            self.setValue(min(inc * steps + self.value(), self.maximum()))
        else:
            self.setValue(max(inc * steps + self.value(), self.minimum()))

    def stepEnabled(self):
        if self.isReadOnly():
            return self.StepEnabledFlag.StepNone
        if self.wrapping():
            return (
                self.StepEnabledFlag.StepDownEnabled
                | self.StepEnabledFlag.StepUpEnabled
            )
        if self.value() < self.maximum():
            if self.value() > self.minimum():
                return (
                    self.StepEnabledFlag.StepDownEnabled
                    | self.StepEnabledFlag.StepUpEnabled
                )
            else:
                return self.StepEnabledFlag.StepUpEnabled
        elif self.value() > self.minimum():
            return self.StepEnabledFlag.StepDownEnabled
        else:
            return self.StepEnabledFlag.StepNone

    def setValue(self, val):
        if np.isnan(val):
            val = np.nan
        else:
            val = max(self.minimum(), min(val, self.maximum()))

        if self._only_int and np.isfinite(val):
            val = round(val)

        self._lastvalue, self._value = self._value, val

        self.valueChanged.emit(self.value())
        self.lineEdit().setText(self.text())
        self.textChanged.emit(self.text())

    def fixup(self, input):
        # Called when the spinbox loses focus with an invalid or intermediate string
        return self.textFromValue(self._lastvalue)

    def validate(self, strn, pos):
        if strn == "-":
            ret = QtGui.QValidator.State.Intermediate
        else:
            ret = QtGui.QValidator.State.Intermediate
            try:
                val = self.valueFromText(strn)
                if val < self.maximum() and val > self.minimum():
                    ret = QtGui.QValidator.State.Acceptable
            except ValueError:
                # sys.excepthook(*sys.exc_info())
                ret = QtGui.QValidator.State.Invalid

        # note: if text is invalid, we don't change the textValid flag since the text
        # will be forced to its previous state anyway
        return (ret, strn, pos)

    def editingFinishedEvent(self):
        self.setValue(self.valueFromText(self.lineEdit().text()))

    def keyPressEvent(self, evt):
        if evt == QtGui.QKeySequence.StandardKey.Copy:
            if (not evt.isAutoRepeat()) and self.lineEdit().hasSelectedText():
                copy_to_clipboard(self.lineEdit().selectedText())
        else:
            super().keyPressEvent(evt)

    def _updateHeight(self):
        if self._is_compact:
            self.setFixedHeight(QtGui.QFontMetrics(self.font()).height() + 3)

    def _get_offset(self):
        spin = QtWidgets.QDoubleSpinBox(self, minimum=0, maximum=0)
        w = (
            spin.minimumSizeHint().width()
            - QtGui.QFontMetrics(spin.font())
            .boundingRect(spin.textFromValue(0.0))
            .width()
        )
        spin.setDisabled(True)
        spin.setVisible(False)
        del spin
        return w + 10

    def _updateWidth(self):
        self.setMinimumWidth(
            max(
                self.widthFromValue(self.maximum()), self.widthFromValue(self.minimum())
            )
            + self._get_offset()
        )


class BetterAxisItem(pg.AxisItem):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def updateAutoSIPrefix(self):
        if self.label.isVisible():
            if self.logMode:
                _range = 10 ** np.array(self.range)
            else:
                _range = self.range
            (scale, prefix) = pg.siScale(
                max(abs(_range[0] * self.scale), abs(_range[1] * self.scale))
            )
            if self.labelUnits == "" and prefix in [
                "k",
                "m",
            ]:  # If we are not showing units, wait until 1e6 before scaling.
                scale = 1.0
                prefix = ""
            self.autoSIPrefixScale = scale
            self.labelUnitPrefix = prefix
        else:
            self.autoSIPrefixScale = 1.0

        self._updateLabel()

    def tickStrings(self, values, scale, spacing):
        if self.logMode:
            return self.logTickStrings(values, scale, spacing)

        places = max(0, np.ceil(-np.log10(spacing * scale)))
        strings = []
        for v in values:
            vs = v * scale
            if abs(vs) < 0.001 or abs(vs) >= 10000:
                vstr = "%g" % vs
            else:
                vstr = ("%%0.%df" % places) % vs
            strings.append(vstr.replace("-", "−"))
        return strings

    def labelString(self):
        if self.labelUnits == "":
            if not self.autoSIPrefix or self.autoSIPrefixScale == 1.0:
                units = ""
            else:
                # units = re.sub(
                #     r"1E\+?(\-?)0?(\d?\d)",
                #     r"10<sup>\1\2</sup>",
                #     f"(×{1.0 / self.autoSIPrefixScale:.3G})",
                # )#.replace("-", "−")

                try:
                    units = "".join(
                        re.search(
                            r"1E\+?(\-?)0?(\d?\d)",
                            f"{1.0 / self.autoSIPrefixScale:.3G}",
                        ).groups()
                    )
                    for k, v in zip(
                        ("0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "-"),
                        ("⁰", "¹", "²", "³", "⁴", "⁵", "⁶", "⁷", "⁸", "⁹", "⁻"),
                    ):
                        units = units.replace(k, v)
                    units = f"10{units}"
                except AttributeError:
                    units = f"{1.0 / self.autoSIPrefixScale:.3G}"
                units = f"(×{units})"

        else:
            units = f"({self.labelUnitPrefix}{self.labelUnits})"

        if self.labelText == "":
            s = units
        else:
            s = f"{self.labelText} {units}"

        style = ";".join([f"{k}: {v}" for k, v in self.labelStyle.items()])

        return f"<span style='{style}'>{s}</span>"

    def setLabel(self, text=None, units=None, unitPrefix=None, **args):
        # `None` input is kept for backward compatibility!
        self.labelText = text or ""
        self.labelUnits = units or ""
        self.labelUnitPrefix = unitPrefix or ""
        if len(args) > 0:
            self.labelStyle = args
        # Account empty string and `None` for units and text
        visible = bool(text or units)
        if text == units == "":
            visible = True
        self.showLabel(visible)
        self._updateLabel()


class FittingParameterWidget(QtWidgets.QWidget):
    sigParamChanged = QtCore.Signal()

    def __init__(
        self,
        name: str,
        spin_kw: dict | None = None,
        checkable: bool = True,
        fixed: bool = False,
        label: str | None = None,
        show_label: bool = True,
    ):
        super().__init__()
        if spin_kw is None:
            spin_kw = {}
        self.layout = QtWidgets.QHBoxLayout(self)
        self.layout.setContentsMargins(0, 0, 0, 0)

        self.param_name = name
        self._prefix = ""
        if label is None:
            label = self.param_name
        self.label = QtWidgets.QLabel(label)
        spin_kw.setdefault("keyboardTracking", False)
        # spin_min_width = spin_kw.pop("minimumWidth", 80)
        self.spin_value = BetterSpinBox(**spin_kw)
        self.spin_lb = BetterSpinBox(
            value=-np.inf,
            minimumWidth=60,
            toolTip="Lower Bound",
            keyboardTracking=False,
        )
        self.spin_ub = BetterSpinBox(
            value=np.inf, minimumWidth=60, toolTip="Upper Bound", keyboardTracking=False
        )
        self.spin_lb.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Fixed
        )
        self.spin_ub.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Fixed
        )
        self.check = QtWidgets.QCheckBox(toolTip="Fixed")

        if show_label:
            self.layout.addWidget(self.label)
        self.layout.addWidget(self.spin_value)
        self.layout.addWidget(self.spin_lb)
        self.layout.addWidget(self.spin_ub)
        self.layout.addWidget(self.check)

        for spin in (self.spin_value, self.spin_lb, self.spin_ub):
            spin.valueChanged.connect(lambda: self.sigParamChanged.emit())
        self.spin_lb.valueChanged.connect(self._refresh_bounds)
        self.spin_ub.valueChanged.connect(self._refresh_bounds)
        self.check.stateChanged.connect(self.setFixed)

        self.setCheckable(checkable)
        self.setFixed(fixed)

    def _refresh_bounds(self):
        self.spin_value.setRange(self.spin_lb.value(), self.spin_ub.value())

    def setValue(self, value):
        self.spin_value.setValue(value)

    def checkable(self):
        return self._checkable

    def setCheckable(self, value: bool):
        self._checkable = value
        self.check.setVisible(value)

    def fixed(self):
        if self.checkable():
            return self.check.isChecked()
        else:
            return False

    def setFixed(self, value: bool):
        if isinstance(value, QtCore.Qt.CheckState):
            if value == QtCore.Qt.CheckState.Unchecked:
                value = False
            elif value == QtCore.Qt.CheckState.Checked:
                value = True
        else:
            self.check.setChecked(value)
        if value:
            self.spin_lb.setDisabled(True)
            self.spin_ub.setDisabled(True)
        else:
            self.spin_lb.setEnabled(True)
            self.spin_ub.setEnabled(True)

    def value(self):
        return self.spin_value.value()

    def prefix(self):
        return self._prefix

    def set_prefix(self, prefix):
        self._prefix = prefix

    def minimum(self):
        return self.spin_lb.value()

    def maximum(self):
        return self.spin_ub.value()

    @property
    def param_dict(self):
        param_info = {"value": self.value()}
        if self.checkable():
            param_info["vary"] = not self.fixed()
        if np.isfinite(self.minimum()):
            param_info["min"] = float(self.minimum())
        if np.isfinite(self.maximum()):
            param_info["max"] = float(self.maximum())
        return {self.prefix() + self.param_name: param_info}


class xImageItem(pg.ImageItem):
    """
    :class:`pyqtgraph.ImageItem` with additional functionality, including
    :class:`xarray.DataArray` support and auto limits based on histogram analysis.

    Parameters
    ----------
    image
        Image data.
    **kwargs
        Additional arguments to :class:`pyqtgraph.ImageItem`.

    Signals
    -------
    sigToleranceChanged()

    """

    sigToleranceChanged = QtCore.Signal(float, float)  #: :meta private:

    def __init__(self, image: npt.NDArray | None = None, **kwargs):
        super().__init__(image, **kwargs)
        self.cut_tolerance = (30, 30)
        self.data_array = None

    def set_cut_tolerance(self, cut_tolerance):
        try:
            self.cut_tolerance = list(cut_tolerance.__iter__)
        except AttributeError:
            self.cut_tolerance = [cut_tolerance] * 2
        self.setImage(levels=self.data_cut_levels())

    def data_cut_levels(self, data=None):
        """Return appropriate levels estimated from the data."""
        if data is None:
            data = self.image
            if data is None:
                return (np.nan, np.nan)
        data = np.nan_to_num(data)
        q3, q1, pu, pl = np.percentile(
            data, [75, 25, 100 - self.cut_tolerance[0], self.cut_tolerance[1]]
        )
        ql, qu = q1 - 1.5 * (q3 - q1), q3 + 1.5 * (q3 - q1)
        self.sigToleranceChanged.emit(
            100 * (data > qu).mean(), 100 * (data < ql).mean()
        )
        mn, mx = max(min(pl, ql), data.min()), min(max(pu, qu), data.max())
        return (mn, mx)

    def setImage(self, image=None, autoLevels=None, cut_to_data=False, **kargs):
        if cut_to_data:
            kargs["levels"] = self.data_cut_levels(data=image)
        super().setImage(image=image, autoLevels=autoLevels, **kargs)

    def setDataArray(self, data=None, **kargs):
        self.data_array = parse_data(data)
        rect = array_rect(self.data_array)
        if self.axisOrder == "row-major":
            img = np.ascontiguousarray(self.data_array.values)
        else:
            img = np.asfortranarray(self.data_array.values.T)
        self.setImage(img, rect=rect, **kargs)


class ParameterGroup(QtWidgets.QGroupBox):
    """Easy creation of groupboxes with multiple varying parameters.

    Can be used in many different interactive tools for dynamic data analysis.

    Parameters
    ----------
    ncols
        Number of columns in the layout.
    groupbox_kw
        Keyword arguments passed onto :class:`PySide6.QtWidgets.QGroupBox`.
    params
        See Examples.


    Signals
    -------
    sigParameterChanged(dict)

    Examples
    --------

    >>> ParameterGroup(
        **{
            "a": QtWidgets.QDoubleSpinBox(range=(0, 1), singleStep=0.01, value=0.2),
            "b": dict(qwtype="dblspin", range=(0, 2), singleStep=0.04),
            "c": QtWidgets.QSlider(range=(0, 10000))
        }
    )

    """

    VALID_QWTYPE: dict[str, QtWidgets.QWidget] = {
        "spin": QtWidgets.QSpinBox,
        "dblspin": QtWidgets.QDoubleSpinBox,
        "btspin": BetterSpinBox,
        "slider": QtWidgets.QSlider,
        "dblslider": QDoubleSlider,
        "chkbox": QtWidgets.QCheckBox,
        "pushbtn": QtWidgets.QPushButton,
        "chkpushbtn": QtWidgets.QPushButton,
        "combobox": QtWidgets.QComboBox,
        "fitparam": FittingParameterWidget,
    }  # : Dictionary of valid widgets that can be added.

    sigParameterChanged: QtCore.SignalInstance = QtCore.Signal(dict)  #: :meta private:

    def __init__(self, ncols: int = 1, groupbox_kw: dict | None = None, **kwargs: dict):
        if groupbox_kw is None:
            groupbox_kw = {}
        super().__init__(**groupbox_kw)
        self.setLayout(QtWidgets.QGridLayout(self))

        self.labels = []
        self.untracked = []
        self.widgets: dict[str, QtWidgets.QWidget] = {}

        j = 0
        for i, (k, v) in enumerate(kwargs.items()):
            if isinstance(v, dict):
                showlabel = v.pop("showlabel", k)
                ind_eff = v.pop("colspan", 1)
                if ind_eff == "ncols":
                    ind_eff = ncols
                if v.pop("notrack", False):
                    self.untracked.append(k)
                self.widgets[k] = self.getParameterWidget(**v)
            elif isinstance(v, QtWidgets.QWidget):
                showlabel = k
                ind_eff = 1
                self.widgets[k] = v
            else:
                raise ValueError(
                    "Each value must be a QtWidgets.QWidget instance"
                    "or a dictionary of keyword arguments to getParameterWidget."
                )

            self.labels.append(QtWidgets.QLabel(str(showlabel)))
            self.labels[i].setBuddy(self.widgets[k])
            if showlabel:
                self.layout().addWidget(self.labels[i], j // ncols, 2 * (j % ncols))
                self.layout().addWidget(
                    self.widgets[k], j // ncols, 2 * (j % ncols) + 1, 1, 2 * ind_eff - 1
                )
            else:
                self.layout().addWidget(
                    self.widgets[k], j // ncols, 2 * (j % ncols), 1, 2 * ind_eff
                )
            j += ind_eff

        self.global_connect()

    @staticmethod
    def getParameterWidget(
        qwtype: (
            Literal[
                "spin",
                "dblspin",
                "btspin",
                "slider",
                "dblslider",
                "chkbox",
                "pushbtn",
                "chkpushbtn",
                "combobox",
                "fitparam",
            ]
            | None
        ) = None,
        **kwargs: dict,
    ):
        """
        Initializes the :class:`PySide6.QtWidgets.QWidget` corresponding to ``qwtype``.

        Parameters
        ----------
        qwtype
            Type of the widget, must a key of :obj:`ParameterGroup.VALID_QWTYPE`.

        """
        if qwtype is None:
            widget = kwargs.pop("widget")
            if not isinstance(widget, QtWidgets.QWidget):
                raise ValueError("widget is not a valid QWidget")
            return widget
        elif qwtype not in ParameterGroup.VALID_QWTYPE:
            raise ValueError(
                f"qwtype must be one of {list(ParameterGroup.VALID_QWTYPE.keys())}"
            )

        widget_class = ParameterGroup.VALID_QWTYPE[qwtype]

        if qwtype == "combobox":
            items = kwargs.pop("items", None)
            currtxt = kwargs.pop("currentText", None)
            curridx = kwargs.pop("currentIndex", None)

        elif qwtype.endswith("pushbtn"):
            pressed = kwargs.pop("pressed", None)
            released = kwargs.pop("released", None)
            if qwtype == "chkpushbtn":
                kwargs["checkable"] = True
                toggled = kwargs.pop("toggled", None)
            else:
                clicked = kwargs.pop("clicked", None)
        newrange = kwargs.pop("range", None)

        valueChanged = kwargs.pop("valueChanged", None)
        textChanged = kwargs.pop("textChanged", None)

        policy = kwargs.pop("policy", None)

        if qwtype == "fitparam":
            show_param_label = kwargs.pop("show_param_label", False)
            kwargs["show_label"] = show_param_label

        fixedWidth = kwargs.pop("fixedWidth", None)
        fixedHeight = kwargs.pop("fixedHeight", None)

        value = kwargs.pop("value", None)

        widget = widget_class(**kwargs)

        if qwtype == "combobox":
            widget.addItems(items)
            if currtxt is not None:
                widget.setCurrentText(currtxt)
            if curridx is not None:
                widget.setCurrentIndex(curridx)
        elif qwtype.endswith("pushbtn"):
            if pressed is not None:
                widget.pressed.connect(pressed)
            if released is not None:
                widget.released.connect(released)
            if qwtype == "chkpushbtn":
                if toggled is not None:
                    widget.toggled.connect(toggled)
            elif clicked is not None:
                widget.clicked.connect(clicked)

        if newrange is not None:
            widget.setRange(*newrange)

        if valueChanged is not None:
            widget.valueChanged.connect(valueChanged)
        if textChanged is not None:
            widget.textChanged.connect(textChanged)

        if fixedWidth is not None:
            widget.setFixedWidth(fixedWidth)
        if fixedHeight is not None:
            widget.setFixedHeight(fixedHeight)
        if policy is not None:
            widget.setSizePolicy(*policy)

        if value is not None:
            widget.setValue(value)

        return widget

    def set_values(self, **kwargs):
        for k, v in kwargs.items():
            self.widgets[k].blockSignals(True)
            self.widgets[k].setValue(v)
            self.widgets[k].blockSignals(False)
        self.sigParameterChanged.emit(kwargs)

    def widget_value(self, widget: str | QtWidgets.QWidget):
        if isinstance(widget, str):
            widget = self.widgets[widget]
        if isinstance(
            widget,
            QtWidgets.QSpinBox
            | QtWidgets.QDoubleSpinBox
            | BetterSpinBox
            | FittingParameterWidget,
        ):
            return widget.value()
        elif isinstance(widget, QtWidgets.QAbstractSpinBox):
            return widget.text()
        elif isinstance(widget, QtWidgets.QAbstractSlider):
            return widget.value()
        elif isinstance(widget, QtWidgets.QCheckBox):
            if widget.isTristate():
                return widget.checkState()
            else:
                return widget.isChecked()
        elif isinstance(widget, QtWidgets.QAbstractButton):
            if widget.isCheckable():
                return widget.isChecked()
            else:
                return widget.isDown()
        elif isinstance(widget, QtWidgets.QComboBox):
            return widget.currentText()

    def widget_change_signal(self, widget):
        if isinstance(
            widget, QtWidgets.QSpinBox | QtWidgets.QDoubleSpinBox | BetterSpinBox
        ):
            return widget.valueChanged
        elif isinstance(widget, FittingParameterWidget):
            return widget.sigParamChanged
        elif isinstance(widget, QtWidgets.QAbstractSpinBox):
            return widget.editingFinished
        elif isinstance(widget, QtWidgets.QAbstractSlider):
            return widget.valueChanged
        elif isinstance(widget, QtWidgets.QCheckBox):
            return widget.stateChanged
        elif isinstance(widget, QtWidgets.QAbstractButton):
            if widget.isCheckable():
                return widget.clicked
            else:
                return widget.toggled
        elif isinstance(widget, QtWidgets.QComboBox):
            return widget.currentTextChanged

    def global_connect(self):
        for k, v in self.widgets.items():
            if k not in self.untracked:
                self.widget_change_signal(v).connect(
                    lambda x=None, name=k: self.sigParameterChanged.emit({name: x})
                )

    def widgets_of_type(self, widgetclass):
        if isinstance(widgetclass, str):
            widgetclass = self.VALID_QWTYPE[widgetclass]
        return [w for w in self.widgets.values() if isinstance(w, widgetclass)]

    @property
    def values(self) -> dict[str, float | int | bool]:
        return {k: self.widget_value(v) for k, v in self.widgets.items()}

    # "spin": QtWidgets.QSpinBox,
    # "dblspin": QtWidgets.QDoubleSpinBox,
    # "slider": QtWidgets.QSlider,
    # "dblslider": QDoubleSlider,
    # "chkbox": QtWidgets.QCheckBox,
    # "pushbtn": QtWidgets.QPushButton,
    # "chkpushbtn": QtWidgets.QPushButton,
    # "combobox": QtWidgets.QComboBox,


class ROIControls(ParameterGroup):
    def __init__(self, roi: pg.ROI, spinbox_kw: dict | None = None, **kwargs):
        if spinbox_kw is None:
            spinbox_kw = {}
        self.roi = roi
        x0, y0, x1, y1 = self.roi_limits
        xm, ym, xM, yM = self.roi.maxBounds.getCoords()

        default_properties = {
            "decimals": 3,
            "singleStep": 0.002,
            "keyboardTracking": False,
        }
        for k, v in default_properties.items():
            spinbox_kw.setdefault(k, v)

        super().__init__(
            x0=dict(
                qwtype="btspin",
                value=x0,
                valueChanged=lambda x: self.modify_roi(x0=x),
                minimum=xm,
                maximum=xM,
                **spinbox_kw,
            ),
            x1=dict(
                qwtype="btspin",
                value=x1,
                valueChanged=lambda x: self.modify_roi(x1=x),
                minimum=xm,
                maximum=xM,
                **spinbox_kw,
            ),
            y0=dict(
                qwtype="btspin",
                value=y0,
                valueChanged=lambda x: self.modify_roi(y0=x),
                minimum=ym,
                maximum=yM,
                **spinbox_kw,
            ),
            y1=dict(
                qwtype="btspin",
                value=y1,
                valueChanged=lambda x: self.modify_roi(y1=x),
                minimum=ym,
                maximum=yM,
                **spinbox_kw,
            ),
            drawbtn={
                "qwtype": "chkpushbtn",
                "toggled": self.draw_mode,
                "showlabel": False,
                "text": "Draw",
            },
            **kwargs,
        )
        self.draw_button = self.widgets["drawbtn"]
        self.roi_spin = [self.widgets[i] for i in ["x0", "y0", "x1", "y1"]]
        self.roi.sigRegionChanged.connect(self.update_pos)

    @property
    def roi_limits(self):
        x0, y0 = self.roi.state["pos"]
        w, h = self.roi.state["size"]
        x1, y1 = x0 + w, y0 + h
        return x0, y0, x1, y1

    def update_pos(self):
        self.widgets["x0"].setMaximum(self.widgets["x1"].value())
        self.widgets["y0"].setMaximum(self.widgets["y1"].value())
        self.widgets["x1"].setMinimum(self.widgets["x0"].value())
        self.widgets["y1"].setMinimum(self.widgets["y0"].value())
        for pos, spin in zip(self.roi_limits, self.roi_spin):
            spin.blockSignals(True)
            spin.setValue(pos)
            spin.blockSignals(False)

    def modify_roi(self, x0=None, y0=None, x1=None, y1=None, update=True):
        lim_new = (x0, y0, x1, y1)
        lim_old = self.roi_limits
        x0, y0, x1, y1 = ((f if f is not None else i) for i, f in zip(lim_old, lim_new))
        xm, ym, xM, yM = self.roi.maxBounds.getCoords()
        x0, y0, x1, y1 = max(x0, xm), max(y0, ym), min(x1, xM), min(y1, yM)
        self.roi.setPos((x0, y0), update=False)
        self.roi.setSize((x1 - x0, y1 - y0), update=update)

    def draw_mode(self, toggle):
        vb = self.roi.parentItem().getViewBox()

        if not toggle:
            vb.mouseDragEvent = self._drag_evt_old
            vb.setMouseMode(self._state_old)
            vb.rbScaleBox.setPen(pg.mkPen((255, 255, 100), width=1))
            vb.rbScaleBox.setBrush(pg.mkBrush(255, 255, 0, 100))
            vb.setCursor(QtGui.QCursor(QtCore.Qt.ArrowCursor))
            return

        vb.setCursor(QtGui.QCursor(QtCore.Qt.CrossCursor))
        self._state_old = vb.state["mouseMode"]
        self._drag_evt_old = vb.mouseDragEvent

        vb.setMouseMode(pg.ViewBox.RectMode)
        vb.rbScaleBox.setPen(pg.mkPen((255, 255, 255), width=1))
        vb.rbScaleBox.setBrush(pg.mkBrush(255, 255, 255, 100))

        def mouseDragEventCustom(ev, axis=None):
            ev.accept()
            pos = ev.pos()
            lastPos = ev.lastPos()
            dif = pos - lastPos

            mouseEnabled = np.array(vb.state["mouseEnabled"], dtype=np.float64)
            mask = mouseEnabled.copy()
            if axis is not None:
                mask[1 - axis] = 0.0

            if ev.button() & QtCore.Qt.MouseButton.LeftButton:
                if vb.state["mouseMode"] == pg.ViewBox.RectMode and axis is None:
                    if ev.isFinish():
                        vb.rbScaleBox.hide()
                        ax = QtCore.QRectF(
                            pg.Point(ev.buttonDownPos(ev.button())), pg.Point(pos)
                        )
                        ax = vb.childGroup.mapRectFromParent(ax)
                        self.modify_roi(*ax.getCoords())
                    else:
                        vb.updateScaleBox(ev.buttonDownPos(), ev.pos())
            elif ev.button() & QtCore.Qt.MouseButton.MiddleButton:
                tr = vb.childGroup.transform()
                tr = pg.invertQTransform(tr)
                tr = tr.map(dif * mask) - tr.map(pg.Point(0, 0))

                x = tr.x() if mask[0] == 1 else None
                y = tr.y() if mask[1] == 1 else None

                vb._resetTarget()
                if x is not None or y is not None:
                    vb.translateBy(x=x, y=y)
                vb.sigRangeChangedManually.emit(vb.state["mouseEnabled"])
            elif ev.button() & QtCore.Qt.MouseButton.RightButton:
                if vb.state["aspectLocked"] is not False:
                    mask[0] = 0

                dif = ev.screenPos() - ev.lastScreenPos()
                dif = np.array([dif.x(), dif.y()])
                dif[0] *= -1
                s = ((mask * 0.02) + 1) ** dif

                tr = pg.invertQTransform(vb.childGroup.transform())
                x = s[0] if mouseEnabled[0] == 1 else None
                y = s[1] if mouseEnabled[1] == 1 else None

                center = pg.Point(
                    tr.map(ev.buttonDownPos(QtCore.Qt.MouseButton.RightButton))
                )
                vb._resetTarget()
                vb.scaleBy(x=x, y=y, center=center)
                vb.sigRangeChangedManually.emit(vb.state["mouseEnabled"])

        vb.mouseDragEvent = mouseDragEventCustom  # set to modified mouseDragEvent


class PostInitCaller(type(QtWidgets.QMainWindow)):
    def __call__(cls, *args, **kwargs):
        obj = type.__call__(cls, *args, **kwargs)
        obj.__post_init__()
        return obj


class AnalysisWindow(QtWidgets.QMainWindow):
    def __init__(
        self,
        data,
        title=None,
        layout="horizontal",
        data_is_input=True,
        analysisWidget=None,
        *args,
        **kwargs,
    ):
        self.qapp = QtCore.QCoreApplication.instance()
        if not self.qapp:
            self.qapp = QtWidgets.QApplication(sys.argv)
        self.qapp.setStyle("Fusion")

        super().__init__()

        self.data = parse_data(data)
        if title is None:
            title = self.data.name
        self.setWindowTitle(title)

        self._main = QtWidgets.QWidget(self)
        self.setCentralWidget(self._main)

        self.controlgroup = QtWidgets.QWidget()
        if layout == "vertical":
            self.layout = QtWidgets.QVBoxLayout(self._main)
            self.controls = QtWidgets.QHBoxLayout(self.controlgroup)
        elif layout == "horizontal":
            self.layout = QtWidgets.QHBoxLayout(self._main)
            self.controls = QtWidgets.QVBoxLayout(self.controlgroup)
        else:
            raise ValueError("Layout must be 'vertical' or 'horizontal'.")

        if analysisWidget is None:
            self.aw = AnalysisWidgetBase(*args, **kwargs)
        elif isinstance(analysisWidget, type):
            self.aw = analysisWidget(*args, **kwargs)
        else:
            self.aw = analysisWidget

        for n in [
            "set_input",
            "add_roi",
            "axes",
            "hists",
            "images",
            # "set_pre_function",
            # "set_pre_function_args",
            # "set_main_function",
            # "set_main_function_args",
            # "refresh_output",
            # "refresh_all",
        ]:
            setattr(self, n, getattr(self.aw, n))
        self.layout.addWidget(self.aw)
        self.layout.addWidget(self.controlgroup)

        self.layout.setContentsMargins(QtCore.QMargins(0, 0, 0, 0))
        self.layout.setSpacing(0)
        self.controlgroup.setContentsMargins(QtCore.QMargins(0, 0, 0, 0))
        if data_is_input:
            self.set_input(data)

    def __post_init__(self, execute=None):
        self.show()
        self.activateWindow()
        self.raise_()

        if execute is None:
            execute = True
            try:
                shell = get_ipython().__class__.__name__  # type: ignore
                if shell in ["ZMQInteractiveShell", "TerminalInteractiveShell"]:
                    execute = False
            except NameError:
                pass
        if execute:
            self.qapp.exec()

    def addParameterGroup(self, *args, **kwargs):
        group = ParameterGroup(*args, **kwargs)
        self.controls.addWidget(group)
        return group

    def closeEvent(self, event: QtGui.QCloseEvent) -> None:
        cb = QtWidgets.QApplication.instance().clipboard()
        if cb.text(cb.Mode.Clipboard) != "":
            pyperclip.copy(cb.text(cb.Mode.Clipboard))
        return super().closeEvent(event)


class AnalysisWidgetBase(pg.GraphicsLayoutWidget):
    """AnalysisWidgetBase

    Parameters
    ----------
    orientation
        Sets the orientation of the plots, by default "vertical"
    num_ax
        Sets the number of axes.
    link
        Link axes, by default "both"
    cut_to_data
        Whether to remove outliers by adjusting color levels, by default "none"

    """

    def __init__(
        self,
        orientation: Literal["vertical", "horizontal"] = "vertical",
        num_ax: int = 2,
        link: Literal["x", "y", "both", "none"] = "both",
        cut_to_data: Literal["in", "out", "both", "none"] = "none",
        **kwargs,
    ):
        super().__init__(**kwargs)
        if orientation == "horizontal":
            self.is_vertical = False
        elif orientation == "vertical":
            self.is_vertical = True
        else:
            raise ValueError("Orientation must be 'vertical' or 'horizontal'.")
        self.cut_to_data = cut_to_data

        self.input = None

        self.initialize_layout(num_ax)

        for i in range(1, num_ax):
            if link in ("x", "both"):
                self.axes[i].setXLink(self.axes[0])
            if link in ("y", "both"):
                self.axes[i].setYLink(self.axes[0])

    def initialize_layout(self, nax):
        self.hists = [pg.HistogramLUTItem() for _ in range(nax)]
        self.axes = [pg.PlotItem() for _ in range(nax)]
        self.images = [xImageItem(axisOrder="row-major") for _ in range(nax)]
        cmap = pg_colormap_powernorm("terrain", 1.0, N=6)
        for i in range(nax):
            self.addItem(self.axes[i], *self.get_axis_pos(i))
            self.addItem(self.hists[i], *self.get_hist_pos(i))
            self.axes[i].addItem(self.images[i])
            self.hists[i].setImageItem(self.images[i])
            self.hists[i].gradient.setColorMap(cmap)
        self.roi = [None for _ in range(nax)]

    def get_axis_pos(self, ax):
        if self.is_vertical:
            return ax, 0, 1, 1
        else:
            return 0, 2 * ax, 1, 1

    def get_hist_pos(self, ax):
        if self.is_vertical:
            return ax, 1, 1, 1
        else:
            return 0, 2 * ax + 1, 1, 1

    def setStretchFactors(self, factors):
        for i, f in enumerate(factors):
            self.setStretchFactor(i, f)

    def setStretchFactor(self, i, factor):
        if self.is_vertical:
            self.ci.layout.setRowStretchFactor(i, factor)
        else:
            self.ci.layout.setColumnStretchFactor(i, factor)

    def set_input(self, data=None):
        if data is not None:
            self.input = parse_data(data)
        self.images[0].setDataArray(
            self.input,
            cut_to_data=self.cut_to_data in ("in", "both"),
        )

    def add_roi(self, i):
        self.roi[i] = pg.ROI(
            [-0.1, -0.5],
            [0.3, 0.5],
            parent=self.images[i],
            rotatable=False,
            resizable=True,
            maxBounds=self.axes[i].getViewBox().itemBoundingRect(self.images[i]),
        )
        self.roi[i].addScaleHandle([0.5, 1], [0.5, 0])
        self.roi[i].addScaleHandle([0.5, 0], [0.5, 1])
        self.roi[i].addScaleHandle([1, 0.5], [0, 0.5])
        self.roi[i].addScaleHandle([0, 0.5], [1, 0.5])
        self.axes[i].addItem(self.roi[i])
        self.roi[i].setZValue(10)
        return self.roi[i]


class ComparisonWidget(AnalysisWidgetBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, num_ax=2, **kwargs)
        self.prefunc = lambda x: x
        self.mainfunc = lambda x: x
        self.prefunc_only_values = False
        self.mainfunc_only_values = False
        self.prefunc_kwargs = {}
        self.mainfunc_kwargs = {}

    def call_prefunc(self, x):
        if self.prefunc_only_values:
            x = np.asarray(x)
        return self.prefunc(x, **self.prefunc_kwargs)

    def set_input(self, data=None):
        if data is not None:
            self.input_ = parse_data(data)

        self.input = self.call_prefunc(self.input_)
        if self.prefunc_only_values:
            self.images[0].setImage(
                np.ascontiguousarray(self.input),
                rect=array_rect(self.input_),
                cut_to_data=self.cut_to_data in ("in", "both"),
            )
        else:
            self.images[0].setDataArray(
                self.input,
                cut_to_data=self.cut_to_data in ("in", "both"),
            )

    def set_pre_function(self, func, only_values=False, **kwargs):
        self.prefunc_only_values = only_values
        self.prefunc = func
        self.set_pre_function_args(**kwargs)
        self.set_input()

    def set_main_function(self, func, only_values=False, **kwargs):
        self.mainfunc_only_values = only_values
        self.mainfunc = func
        self.set_main_function_args(**kwargs)
        self.refresh_output()

    def set_pre_function_args(self, **kwargs):
        for k, v in kwargs.items():
            self.prefunc_kwargs[k] = v
        self.refresh_output()

    def set_main_function_args(self, **kwargs):
        for k, v in kwargs.items():
            self.mainfunc_kwargs[k] = v
        self.refresh_output()

    def refresh_output(self):
        if self.mainfunc_only_values:
            self.output = self.mainfunc(np.asarray(self.input_), **self.mainfunc_kwargs)
            self.images[1].setImage(
                np.ascontiguousarray(self.output),
                rect=array_rect(self.input),
                cut_to_data=self.cut_to_data in ("out", "both"),
            )
        else:
            self.output = self.mainfunc(self.input_, **self.mainfunc_kwargs)
            self.images[1].setDataArray(
                self.output,
                cut_to_data=self.cut_to_data in ("out", "both"),
            )

    def refresh_all(self):
        self.set_input()
        self.refresh_output()


class DictMenuBar(QtWidgets.QMenuBar):
    def __init__(self, parent: QtWidgets.QWidget | None = ..., **kwargs) -> None:
        super().__init__(parent)

        self.menu_dict: dict[str, QtWidgets.QMenu] = {}
        self.action_dict: dict[str, QtWidgets.QAction] = {}

        self.add_items(**kwargs)

    def __getattribute__(self, __name: str) -> Any:
        try:
            return super().__getattribute__(__name)
        except AttributeError:
            try:
                out = self.menu_dict[__name]
            except KeyError:
                out = self.action_dict[__name]
            warnings.warn(
                f"Menu or Action '{__name}' called as an attribute",
                PendingDeprecationWarning,
                stacklevel=2,
            )
            return out

    def add_items(self, **kwargs):
        self.parse_menu(self, **kwargs)

    def parse_menu(self, parent: QtWidgets.QMenuBar | QtWidgets.QMenu, **kwargs: dict):
        for name, opts in kwargs.items():
            menu = opts.pop("menu", None)
            actions = opts.pop("actions")

            if menu is None:
                title = opts.pop("title", None)
                icon = opts.pop("icon", None)
                if title is None:
                    title = name
                if icon is None:
                    menu = parent.addMenu(title)
                else:
                    menu = parent.addMenu(icon, title)
            else:
                menu = parent.addMenu(menu)

            self.menu_dict[name] = menu

            for actname, actopts in actions.items():
                if isinstance(actopts, QtWidgets.QAction):
                    act = actopts
                    sep_before, sep_after = False, False
                else:
                    if "actions" in actopts:
                        self.parse_menu(menu, **{actname: actopts})
                        continue
                    sep_before = actopts.pop("sep_before", False)
                    sep_after = actopts.pop("sep_after", False)
                    if "text" not in actopts:
                        actopts["text"] = actname
                    act = self.parse_action(actopts)
                if sep_before:
                    menu.addSeparator()
                menu.addAction(act)
                if (
                    act.text() is not None
                ):  # check whether it's a separator without text
                    self.action_dict[actname] = act
                if sep_after:
                    menu.addSeparator()

    @staticmethod
    def parse_action(actopts: dict):
        shortcut = actopts.pop("shortcut", None)
        triggered = actopts.pop("triggered", None)
        toggled = actopts.pop("toggled", None)
        changed = actopts.pop("changed", None)

        if shortcut is not None:
            actopts["shortcut"] = QtGui.QKeySequence(shortcut)

        action = QtGui.QAction(**actopts)

        if triggered is not None:
            action.triggered.connect(triggered)
        if toggled is not None:
            action.toggled.connect(toggled)
        if changed is not None:
            action.changed.connect(changed)
        return action


if __name__ == "__main__":
    from scipy.ndimage import gaussian_filter  # , uniform_filter

    qapp = QtWidgets.QApplication.instance()
    if not qapp:
        qapp = QtWidgets.QApplication(sys.argv)
    qapp.setStyle("Fusion")

    dat = (
        xr.open_dataarray(
            "/Users/khan/Documents/ERLab/CsV3Sb5/2021_Dec_ALS_CV3Sb5/Data/cvs_kxy_small.nc"
        )
        .sel(eV=-0.15, method="nearest")
        .fillna(0)
    )
    win = AnalysisWindow(dat, analysisWidget=ComparisonWidget, orientation="vertical")
    win.setAttribute(QtCore.Qt.WidgetAttribute.WA_AcceptTouchEvents, False)

    def gaussfilt_2d(dat, sx, sy):
        return gaussian_filter(dat, sigma=(sx, sy))

    win.aw.set_main_function(gaussfilt_2d, sx=0.1, sy=1, only_values=True)

    # win.set_pre_function(gaussian_filter, sigma=[1, 1], only_values=True)
    # win.set_pre_function(gaussian_filter, sigma=(0.1, 0.1))

    # layout.addWidget(win)
    win.addParameterGroup(
        sigma_x={
            "qwtype": "btspin",
            "minimum": 0,
            "maximum": 10,
            "valueChanged": lambda x: win.aw.set_main_function_args(sx=x),
        },
        sigma_y={
            "qwtype": "btspin",
            "minimum": 0,
            "maximum": 10,
            "valueChanged": lambda x: win.aw.set_main_function_args(sy=x),
        },
        b={"qwtype": "combobox", "items": ["item1", "item2", "item3"]},
    )

    win.__post_init__(execute=True)
    # new_roi = win.add_roi(0)

    # layout.addWidget(ROIControls(new_roi))

    # wdgt.show()
    # wdgt.activateWindow()
    # wdgt.raise_()



"""Macros for correlation analysis."""

__all__ = ["acf2stack", "acf2", "match_dims", "xcorr1d"]


import itertools

import joblib
import numpy as np
import scipy.signal
import xarray as xr


def autocorrelate(arr, *args, **kwargs):
    """Calculate the autocorrelation of a N-dimensional array, normalized to 1.

    Parameters
    ----------
    arr
        Input array to calculate the autocorrelation.
    *args, **kwargs
        Additional arguments and keyword arguments to be passed to
        `scipy.signal.correlate`.

    Returns
    -------
    autocorr
        Autocorrelation of the input array.

    """
    acf = scipy.signal.correlate(arr, arr, *args, **kwargs)
    return acf / acf[tuple(s // 2 for s in acf.shape)]


def autocorrelation_lags(in_len, *args, **kwargs):
    return scipy.signal.correlation_lags(in_len, in_len, *args, **kwargs)


def nanacf(arr, *args, **kwargs):
    acf = autocorrelate(np.nan_to_num(arr), *args, **kwargs)
    if np.isnan(arr).any():
        nan_mask = ~np.isnan(arr)
        acf_nan = autocorrelate(nan_mask.astype(float), *args, **kwargs)
        acf_nan[acf_nan < 1e7 * np.finfo(float).eps] = np.nan
        return acf / acf_nan
    else:
        return acf


def acf2(arr, mode: str = "full", method: str = "fft"):
    """
    Calculate the autocorrelation function (ACF) of a two-dimensional array including
    nan values.

    Parameters
    ----------
    arr
        The input array for which the ACF needs to be calculated.
    mode
        The mode of the ACF calculation, by default ``"full"``. For more information,
        see `scipy.signal.correlate`.
    method
        The method used for ACF calculation, by default ``"fft"``. For more information,
        see `scipy.signal.correlate`.

    Returns
    -------
    xarray.DataArray
        The ACF of the input array.

    Examples
    --------
    >>> import numpy as np
    >>> import xarray as xr
    >>> np.random.seed(0)  # Set the random seed for reproducibility
    >>> arr = xr.DataArray(np.random.rand(10, 10), dims=("kx", "ky"))
    >>> acf = acf2(arr)
    >>> acf
    <xarray.DataArray (qx: 19, qy: 19)> Size: 3kB
    8.403e-05 0.01495 0.01979 0.02734 0.03215 ... 0.02734 0.01979 0.01495 8.403e-05
    Coordinates:
    * qx       (qx) int64 152B -9 -8 -7 -6 -5 -4 -3 -2 -1 0 1 2 3 4 5 6 7 8 9
    * qy       (qy) int64 152B -9 -8 -7 -6 -5 -4 -3 -2 -1 0 1 2 3 4 5 6 7 8 9

    """
    out = arr.copy(deep=True)
    acf = nanacf(out.values, mode=mode, method=method)
    coords = [out[d].values for d in out.dims]
    steps = [c[1] - c[0] for c in coords]
    out = xr.DataArray(
        acf,
        {
            d: autocorrelation_lags(n, mode) * s
            for s, n, d in zip(steps, arr.shape, out.dims)
        },
        attrs=out.attrs,
    )
    if all(i in out.dims for i in ["kx", "ky"]):
        out = out.rename({"kx": "qx", "ky": "qy"})
    return out


def acf2stack(arr, stack_dims=("eV",), mode: str = "full", method: str = "fft"):
    if arr.ndim == 2:
        return acf2(arr, mode, method)
    elif arr.ndim >= 3:
        if arr.ndim - len(stack_dims) != 2:
            raise ValueError(
                "The number of dimensions excluding the stacking dimensions must be 2"
            )

        stack_sizes = {d: len(arr[d]) for d in stack_dims}
        stack_iter = tuple(range(s) for s in stack_sizes.values())

        out_list = joblib.Parallel(n_jobs=-1, pre_dispatch="3 * n_jobs")(
            joblib.delayed(nanacf)(
                np.squeeze(arr.isel(dict(zip(stack_dims, vals))).values),
                mode,
                method,
            )
            for vals in itertools.product(*stack_iter)
        )
        acf_dims = tuple(filter(lambda d: d not in stack_dims, arr.dims))
        acf_sizes = dict(zip(acf_dims, out_list[0].shape))
        acf_steps = tuple(arr[d].values[1] - arr[d].values[0] for d in acf_dims)

        out_sizes = stack_sizes | acf_sizes

        if mode == "same":
            out = arr.copy(deep=True)
        else:
            out = xr.DataArray(
                np.empty(tuple(out_sizes[d] for d in arr.dims)),
                dims=arr.dims,
                attrs=arr.attrs,
            )
            out = out.assign_coords({d: arr[d] for d in stack_dims})

        for i, vals in enumerate(itertools.product(*stack_iter)):
            out.loc[{s: arr[s][v] for s, v in zip(stack_dims, vals)}] = out_list[i]

        out = out.assign_coords(
            {
                d: autocorrelation_lags(len(arr[d]), mode) * s
                for s, d in zip(acf_steps, acf_dims)
            }
        )
        if all(i in out.dims for i in ["kx", "ky"]):
            out = out.rename({"kx": "qx", "ky": "qy"})
    return out


def match_dims(da1: xr.DataArray, da2: xr.DataArray):
    """
    Returns the second array interpolated with the coordinates of the first array,
    making them the same size.

    """
    return da2.interp({dim: da1[dim] for dim in da2.dims})


def xcorr1d(in1: xr.DataArray, in2: xr.DataArray, method="direct"):
    """Performs 1-dimensional correlation analysis on `xarray.DataArray` s."""
    in2 = match_dims(in1, in2)
    out = in1.copy(deep=True)
    xind = scipy.signal.correlation_lags(in1.values.size, in2.values.size, mode="same")
    xzero = np.flatnonzero(xind == 0)[0]
    out.values = scipy.signal.correlate(
        in1.fillna(0).values, in2.fillna(0).values, mode="same", method=method
    )
    out[in1.dims[0]] = out[in1.dims[0]] - out[in1.dims[0]][xzero]
    return out
